# cython: language_level=3
# cython: boundscheck=True
# cython: wraparound=False
# cython: cdivision=True
#
################################################################################
# Utility functions for Ray-tracing
################################################################################

# ==============================================================================
# =  3D Bounding box (not Toroidal)
# ==============================================================================
cdef void compute_3d_bboxes(const double*const* vignett_poly,
                            const long* lnvert,
                            const int nvign,
                            double* lbounds,
                            const int num_threads) nogil
cdef void comp_bbox_poly3d(const int nvert,
                           const double* vertx,
                           const double* verty,
                           const double* vertz,
                           double[6] bounds) nogil

# ==============================================================================
# =  Computation of Bounding Boxes (in toroidal configuration)
# ==============================================================================
cdef void comp_bbox_poly_tor(const int nvert,
                             const double* vertr,
                             const double* vertz,
                             double[6] bounds) nogil

cdef void comp_bbox_poly_tor_lim(const int nvert,
                                 const double* vertr,
                                 const double* vertz,
                                 double[6] bounds,
                                 const double lmin,
                                 const double lmax) nogil

cdef void coordshift_simple1d(double[3] pts, bint in_is_cartesian,
                              double CrossRef, double cos_phi,
                              double sin_phi) nogil

cdef void is_inside_vessel(double[:, ::1] pts, double[:, ::1] ves_poly,
                           double[:,::1] ves_lims, int nlim,
                           bint is_toroidal, bint in_is_cartesian,
                           int[3] order,
                           int[::1] is_inside) nogil

# ==============================================================================
# =  Raytracing basic tools: intersection ray and axis aligned bounding box
# ==============================================================================
cdef bint inter_ray_aabb_box(const int[3] sign,
                             const double[3] inv_direction,
                             const double* bounds,
                             const double[3] ds,
                             const bint countin) nogil

# ==============================================================================
# =  Raytracing basic tools: intersection ray and triangle (in 3d space)
# ==============================================================================
cdef bint inter_ray_triangle(const double[3] ray_orig,
                             const double[3] ray_vdir,
                             const double* vert0,
                             const double* vert1,
                             const double* vert2) nogil

# ==============================================================================
# =  Raytracing on a Torus
# ==============================================================================
cdef void raytracing_inout_struct_tor(const int num_los,
                                      const double[:,::1] ray_vdir,
                                      const double[:,::1] ray_orig,
                                      double[::1] coeff_inter_out,
                                      double[::1] coeff_inter_in,
                                      double[::1] vperp_out,
                                      const long* lstruct_nlim,
                                      int[::1] ind_inter_out,
                                      const bint forbid0,
                                      const bint forbidbis,
                                      const double rmin,
                                      const double rmin2,
                                      const double crit2_base,
                                      const int nstruct_lim,
                                      const double* lbounds,
                                      const double* langles,
                                      const int* lis_limited,
                                      const long* lnvert,
                                      const long* lsz_lim,
                                      const double* lstruct_polyx,
                                      const double* lstruct_polyy,
                                      const double* lstruct_normx,
                                      const double* lstruct_normy,
                                      const double eps_uz,
                                      const double eps_vz,
                                      const double eps_a,
                                      const double eps_b,
                                      const double eps_plane,
                                      const int num_threads,
                                      const bint is_out_struct) nogil

# ------------------------------------------------------------------
cdef bint comp_inter_los_vpoly(const double[3] ray_orig,
                               const double[3] ray_vdir,
                               const double* lpolyx,
                               const double* lpolyy,
                               const double* normx,
                               const double* normy,
                               const int nvert,
                               const bint lim_is_none,
                               const double lim_min,
                               const double lim_max,
                               const bint forbidbis,
                               const double upscaDp, const double upar2,
                               const double dpar2, const double invuz,
                               const double s1x,   const double s1y,
                               const double s2x, const double s2y,
                               const double crit2, const double eps_uz,
                               const double eps_vz, const double eps_a,
                               const double eps_b, const double eps_pln,
                               const bint is_in_struct,
                               double[1] kpin_loc, double[1] kpout_loc,
                               int[1] ind_loc, double[3] vperpin) nogil

# ==============================================================================
# =  Raytracing on a Cylinder (Linear case)
# ==============================================================================
cdef void raytracing_inout_struct_lin(const int Nl,
                                      const double[:,::1] Ds,
                                      const double [:,::1] us,
                                      const int Ns,
                                      const double* polyx_tab,
                                      const double* polyy_tab,
                                      const double* normx_tab,
                                      const double* normy_tab,
                                      const double L0, const double L1,
                                      double[::1] kin_tab,
                                      double[::1] kout_tab,
                                      double[::1] vperpout_tab,
                                      int[::1] indout_tab,
                                      const double EpsPlane,
                                      const int ind_struct,
                                      const int ind_lim_struct) nogil

cdef void compute_inout_tot(const int nl, const int np,
                            const double[:, ::1] ray_orig,
                            const double[:, ::1] ray_vdir,
                            const double[:, ::1] ves_poly,
                            const double[:, ::1] ves_norm,
                            long[::1] lstruct_nlim,
                            const double[::1] ves_lims,
                            const double[::1] lstruct_polyx,
                            const double[::1] lstruct_polyy,
                            const double[::1] lstruct_lims,
                            const double[::1] lstruct_normx,
                            const double[::1] lstruct_normy,
                            const long[::1] lnvert,
                            const int nstruct_tot,
                            const int nstruct_lim,
                            const int sz_ves_lims,
                            const double min_poly_r,
                            const double rmin,
                            const double eps_uz, const double eps_a,
                            const double eps_vz, const double eps_b,
                            const double eps_plane, bint is_tor,
                            const bint forbid, const int num_threads,
                            double[::1] coeff_inter_out,
                            double[::1] coeff_inter_in,
                            double[::1] vperp_out,
                            int[::1] ind_inter_out) nogil

# ==============================================================================
# =  Raytracing on a Torus only KMin and KMax
# ==============================================================================
cdef void raytracing_minmax_struct_tor(const int num_los,
                                       const double[:,::1] ray_vdir,
                                       const double[:,::1] ray_orig,
                                       double* coeff_inter_out,
                                       double* coeff_inter_in,
                                       const bint forbid0,
                                       const bint forbidbis,
                                       const double rmin,
                                       const double rmin2,
                                       const double crit2_base,
                                       const int npts_poly,
                                       const double* langles,
                                       const bint is_limited,
                                       const double* surf_polyx,
                                       const double* surf_polyy,
                                       const double* surf_normx,
                                       const double* surf_normy,
                                       const double eps_uz,
                                       const double eps_vz,
                                       const double eps_a,
                                       const double eps_b,
                                       const double eps_plane,
                                       const int num_threads) nogil

# ==============================================================================
# =  Raytracing on a Cylinder only KMin and KMax
# ==============================================================================
cdef void raytracing_minmax_struct_lin(const int Nl,
                                       const double[:,::1] Ds,
                                       const double [:,::1] us,
                                       const int Ns,
                                       const double* polyx_tab,
                                       const double* polyy_tab,
                                       const double* normx_tab,
                                       const double* normy_tab,
                                       const double L0,
                                       const double L1,
                                       double* kin_tab,
                                       double* kout_tab,
                                       const double EpsPlane) nogil

# ==============================================================================
# = Checking if points are visible
# ==============================================================================
cdef void is_visible_pt_vec(double pt0, double pt1, double pt2,
                            double[:,::1] pts, int npts,
                            double[:, ::1] ves_poly,
                            double[:, ::1] ves_norm,
                            long* is_vis,
                            double[::1] k,
                            double[::1] ves_lims,
                            long[::1] lstruct_nlim,
                            double[::1] lstruct_polyx,
                            double[::1] lstruct_polyy,
                            double[::1] lstruct_lims,
                            double[::1] lstruct_normx,
                            double[::1] lstruct_normy,
                            long[::1] lnvert,
                            int nstruct_tot,
                            int nstruct_lim,
                            double rmin,
                            double eps_uz, double eps_a,
                            double eps_vz, double eps_b,
                            double eps_plane, bint is_tor,
                            bint forbid,
                            int num_threads)


cdef void are_visible_vec_vec(double[:, ::1] pts1, int npts1,
                              double[:,::1] pts2, int npts2,
                              double[:, ::1] ves_poly,
                              double[:, ::1] ves_norm,
                              long[:, ::1] ind,
                              double[:, ::1] dist,
                              double[::1] ves_lims,
                              long[::1] lstruct_nlim,
                              double[::1] lstruct_polyx,
                              double[::1] lstruct_polyy,
                              double[::1] lstruct_lims,
                              double[::1] lstruct_normx,
                              double[::1] lstruct_normy,
                              long[::1] lnvert,
                              int nstruct_tot,
                              int nstruct_lim,
                              double rmin,
                              double eps_uz, double eps_a,
                              double eps_vz, double eps_b,
                              double eps_plane, bint is_tor,
                              bint forbid,
                              int num_threads)

cdef void is_visible_pt_vec_core(double pt0, double pt1, double pt2,
                                 double[:, ::1] pts, int npts,
                                 double[:, ::1] ves_poly,
                                 double[:, ::1] ves_norm,
                                 long* is_vis,
                                 double* dist,
                                 double[::1] ves_lims,
                                 long[::1] lstruct_nlim,
                                 double[::1] lstruct_polyx,
                                 double[::1] lstruct_polyy,
                                 double[::1] lstruct_lims,
                                 double[::1] lstruct_normx,
                                 double[::1] lstruct_normy,
                                 long[::1] lnvert,
                                 # results:
                                 double[::1] vperp_out,
                                 double[::1] coeff_inter_in,
                                 double[::1] coeff_inter_out,
                                 int[::1] ind_inter_out,
                                 int sz_ves_lims,
                                 double[:, ::1] ray_orig,
                                 double[:, ::1] ray_vdir,
                                 int npts_poly,
                                 # ...
                                 int nstruct_tot,
                                 int nstruct_lim,
                                 double rmin,
                                 double eps_uz, double eps_a,
                                 double eps_vz, double eps_b,
                                 double eps_plane, bint is_tor,
                                 bint forbid,
                                 int num_threads) nogil

cdef void is_visible_pt_vec_core_nd(double pt0, double pt1, double pt2,
                                    double[:, ::1] pts, int npts,
                                    double[:, ::1] ves_poly,
                                    double[:, ::1] ves_norm,
                                    long* is_vis,
                                    double[::1] ves_lims,
                                    long[::1] lstruct_nlim,
                                    double[::1] lstruct_polyx,
                                    double[::1] lstruct_polyy,
                                    double[::1] lstruct_lims,
                                    double[::1] lstruct_normx,
                                    double[::1] lstruct_normy,
                                    long[::1] lnvert,
                                    # results:
                                    double[::1] vperp_out,
                                    double[::1] coeff_inter_in,
                                    double[::1] coeff_inter_out,
                                    int[::1] ind_inter_out,
                                    int sz_ves_lims,
                                    double[:, ::1] ray_orig,
                                    double[:, ::1] ray_vdir,
                                    int npts_poly,
                                    # ...
                                    int nstruct_tot,
                                    int nstruct_lim,
                                    double rmin,
                                    double eps_uz, double eps_a,
                                    double eps_vz, double eps_b,
                                    double eps_plane, bint is_tor,
                                    bint forbid,
                                    int num_threads) nogil



# -*- coding: utf-8 -*-


import numpy as np
import scipy.constants as scpct
import datastock as ds


from . import _generic_check
from ..spectro import _rockingcurve_def


_DMAT_KEYS = {
    'name': {'types': str},
    'symbol': {'types': str},
    'd_hkl': {'types': float, 'sign': '> 0.'},
    'target_ion': {'types': str},
    'target_lamb': {'types': float, 'sign': '> 0.'},
    'atoms': {'types': (list, np.ndarray), 'types_iter': str},
    'atoms_Z': {'dtype': int, 'size': 2, 'sign': '> 0'},
    'atoms_nb': {'dtype': int, 'size': 2, 'sign': '> 0'},
    'miller': {'dtype': int, 'size': 3, 'sign': '>= 0'},
    'alpha': {'types': float, 'default': 0., 'sign': '>= 0'},
    'beta': {'types': float, 'default': 0.},
    'mesh': {'types': dict},
    'phases': {'types': dict},
    'inter_atomic': {'types': dict},
    'thermal_expansion': {'types': dict},
    'atomic_scattering': {'types': dict},
    'sin_theta_lambda': {'types': dict},
}


# #############################################################################
# #############################################################################
#                           Crystal
# #############################################################################


def _dmat(
    dmat=None,
    alpha=None,
    beta=None,
    dgeom=None,
):

    # ---------------------
    # Easy cases
    # ---------------------

    # not mat
    if dmat is None:
        return dmat

    # known crystal
    ready_to_compute = False
    if isinstance(dmat, str):
        if dmat not in _rockingcurve_def._DCRYST.keys():
            msg = (
                f"Arg dmat points to an unknown crystal: '{dmat}'"
            )
            raise Exception(msg)
        dmat = _rockingcurve_def._DCRYST[dmat]
        ready_to_compute = True

    # ---------------------
    # check dict integrity
    # ---------------------

    # Check dict typeand content (each key is a valid string)
    dmat = ds._generic_check._check_dict_valid_keys(
        var=dmat,
        varname='dmat',
        has_all_keys=False,
        has_only_keys=False,
        keys_can_be_None=True,
        dkeys=_DMAT_KEYS,
    )

    dmat['ready_to_compute'] = ready_to_compute

    # -------------------------------
    # check each value independently
    # -------------------------------

    # target ion

    # alpha
    if alpha is not None:
        dmat['alpha'] = alpha
    if dmat['alpha'] is not None:
        dmat['alpha'] = np.abs(np.arctan(
            np.sin(dmat['alpha']),
            np.cos(dmat['alpha']),
        ))

    # beta
    if beta is not None:
        dmat['beta'] = beta
    if dmat['beta'] is not None:
        dmat['beta'] = np.arctan2(np.sin(dmat['beta']), np.cos(dmat['beta']))

    # vector basis with non-paralellism
    if all([dmat[k0] is not None for k0 in ['alpha', 'beta']]):
        nin = (
            np.cos(dmat['alpha'])*(dgeom['nin'])
            + np.sin(dmat['alpha']) * (
                np.cos(dmat['beta'])*dgeom['e0']
                + np.sin(dmat['beta'])*dgeom['e1']
            )
        )
        e0 = (
            - np.sin(dmat['alpha'])*(dgeom['nin'])
            + np.cos(dmat['alpha']) * (
                np.cos(dmat['beta'])*dgeom['e0']
                + np.sin(dmat['beta'])*dgeom['e1']
            )
        )
        nin, e0, e1 = ds._generic_check._check_vectbasis(
            e0=nin,
            e1=e0,
            e2=None,
            ndim=3,
        )

        dmat['nin'] = nin
        dmat['e0'] = e0
        dmat['e1'] = e1

    # -------------------------------
    # check sub-dict
    # -------------------------------

    ldict = [k0 for k0, v0 in _DMAT_KEYS.items() if v0.get('types') is dict]
    for k0 in ldict:
        pass
    # TODO: implement further checks of sub-dict ?

    return dmat




"""
This module contains tests for tofu.geom in its structured version
"""


# Built-in
import os
import warnings


# Standard
import numpy as np
import scipy.constants as scpct
import matplotlib.pyplot as plt


# tofu-specific
from tofu import __version__
import tofu.data as tfd
import tofu.utils as tfu

_here = os.path.abspath(os.path.dirname(__file__))
VerbHead = 'tofu.data.SpectralLines'


#######################################################
#
#     Setup and Teardown
#
#######################################################


def setup_module(module):
    print("")   # this is to get a newline after the dots
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if not lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following previous test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    # print("setup_module before anything in this file")


def teardown_module(module):
    # os.remove(VesTor.Id.SavePath + VesTor.Id.SaveName + '.npz')
    # os.remove(VesLin.Id.SavePath + VesLin.Id.SaveName + '.npz')
    # print("teardown_module after everything in this file")
    # print("") # this is to get a newline
    LF = os.listdir(_here)
    lss = ['TFD_', 'Test', '.npz']
    LF = [lf for lf in LF if all([ss in lf for ss in lss])]
    LF = [
        lf for lf in LF
        if lf[lf.index('_Vv')+2:lf.index('_U')] == __version__
    ]
    print("Removing the following test files:")
    print(LF)
    for lf in LF:
        os.remove(os.path.join(_here, lf))
    pass


# #############################################################################
# #############################################################################
#           Specific to SpectralLines
# #############################################################################


class Test01_SpectralLines(object):

    @classmethod
    def setup_class(cls, Name='data1',  SavePath='./', verb=False):
        cls.sl = tfd.SpectralLines.from_openadas(
            lambmin=3.94e-10,
            lambmax=4e-10,
            element=['Ar', 'W'],
        )

    @classmethod
    def setup_method(self):
        pass

    def teardown_method(self):
        pass

    @classmethod
    def teardown_class(cls):
        pass

    def test01_add_from_openadas(self):
        lines = self.sl.dobj['lines']
        self.sl.add_from_openadas(
            lambmin=3.90e-10,
            lambmax=3.96e-10,
            element='W',
        )
        assert all([k0 in self.sl.dobj['lines'].keys() for k0 in lines.keys()])

    def test02_sortby(self):
        self.sl.sortby(param='lambda0', which='lines')
        self.sl.sortby(param='ion', which='lines')

    def test03_convert_lines(self):
        self.sl.convert_lines(units='Hz')

    def test04_calc_pec(self):
        ne = np.r_[1e15, 1e18, 1e21]
        Te = np.r_[1e3, 2e3, 3e3, 4e3, 5e3]
        dpec = self.sl.calc_pec(ne=ne, Te=Te[:ne.size], grid=False)
        dpec = self.sl.calc_pec(
            key='Ar16_9_oa_pec40_cl', ne=ne, Te=Te[:ne.size], grid=False,
        )
        dpec = self.sl.calc_pec(ne=ne, Te=Te, grid=True)
        dpec = self.sl.calc_pec(
            key='Ar16_9_oa_pec40_cl', ne=ne, Te=Te[:ne.size], grid=False,
        )

    def test05_calc_intensity(self):
        ne = np.r_[1e15, 1e18, 1e21]
        Te = np.r_[1e3, 2e3, 3e3, 4e3, 5e3]

        concentration = np.r_[0.1, 0.2, 0.3]
        dint = self.sl.calc_intensity(
            ne=ne, Te=Te[:ne.size], concentration=concentration, grid=False,
        )

        key = ['Ar16_9_oa_pec40_cl']
        concentration = {k0: np.r_[0.1, 0.2, 0.3] for k0 in key}
        dint = self.sl.calc_intensity(
            key=key,
            ne=ne, Te=Te[:ne.size], concentration=concentration, grid=False,
        )

        concentration = np.random.random((ne.size, Te.size))
        dint = self.sl.calc_intensity(
            ne=ne, Te=Te, concentration=concentration, grid=True,
        )

        key = ['Ar16_9_oa_pec40_cl']
        concentration = {k0: concentration for k0 in key}
        dint = self.sl.calc_intensity(
            key=key,
            ne=ne, Te=Te, concentration=concentration, grid=True,
        )

    def test06_plot_spectral_lines(self):
        ax = self.sl.plot_spectral_lines()
        plt.close('all')

    def test07_plot_pec_single(self):
        Te = 1.e3
        ne = 1.e20
        ax = self.sl.plot_pec_single(Te=Te, ne=ne)

    # def test08_plot_pec(self):
        # Te = np.linspace(1, 7, 7)*1e3
        # ne = np.logspace(15, 21, 7)
        # ax = self.sl.plot_pec(Te=1e3, ne=ne)
        # ax = self.sl.plot_pec(Te=Te, ne=1e19)
        # ax = self.sl.plot_pec(Te=Te, ne=ne)
        # plt.close('all')



# -*- coding: utf-8 -*-


import numpy as np
import scipy.constants as scpct
import datastock as ds


_DMAT_KEYS = {
    'name': {'types': str},
    'symbol': {'types': str},
    'thickness': {'types': float, 'sign': '> 0.'},
    'energy': {'dtype': float, 'sign': '> 0.'},
    'transmission': {'dtype': float, 'sign': ['>= 0.', '<= 1.']},
}


# #############################################################################
# #############################################################################
#                       dmat for filter
# #############################################################################


def _trim_1d(y=None, val=None, trim=None):

    # check input
    if trim is None:
        trim = 'fb'

    if 'f' not in trim and 'b' not in trim:
        return

    # compute
    iv = (y == val).nonzero()[0]

    ind = None
    if iv.size > 0:

        if 'f' in trim:
            i0 = np.array([
                ii for ii in range(iv[-1])
                if ii in iv and ii + 1 in iv
            ])
            if i0.size > 0:
                ind = i0

        if 'b' in trim:
            i1 = np.sort([
                ii for ii in range(y.size - 1, iv[0] - 1, -1)
                if ii in iv and ii - 1 in iv
            ])

            if i1.size > 0:
                if ind is None:
                    ind = i1
                else:
                    ind = np.unique(np.r_[ind, i1])

    return ind


def _dmat_energy_trans(energ=None, trans=None):

    if energ.size != trans.size:
        msg = (
            f"The following should be 1d arrays of the same size:\n"
            f"\t- dmat['energy'].shape = {dmat['energy'].shape}\n"
            f"\t- dmat['transmission'].shape = {dmat['transmission'].shape}\n"
        )
        raise Exception(msg)

    # make sure all values are finite
    iok = np.isfinite(energ) & np.isfinite(trans)
    energ = energ[iok]
    trans = trans[iok]

    # make sure energy os sorted and unique
    energ, inds = np.unique(energ, return_index=True)
    trans = trans[inds]

    # remove trailing 0 and 1 in trans
    ind0 = _trim_1d(y=trans, val=0, trim='fb')
    ind1 = _trim_1d(y=trans, val=1, trim='fb')

    ind = None
    if ind0 is not None:
        ind = ind0
    if ind1 is not None:
        if ind0 is None:
            ind = ind1
        else:
            ind = np.r_[ind0, ind1]

    if ind is not None:
        energ = np.delete(energ, np.r_[ind0, ind1])
        trans = np.delete(trans, np.r_[ind0, ind1])

    return energ, trans


def _dmat(
    key=None,
    dmat=None,
):

    # ---------------------
    # Easy cases
    # ---------------------

    # not mat
    if dmat is None:
        return None, None, None

    # ---------------------
    # check dict integrity
    # ---------------------

    # Check dict type and content (each key is a valid string)
    dmat = ds._generic_check._check_dict_valid_keys(
        var=dmat,
        varname='dmat',
        has_all_keys=True,
        has_only_keys=False,
        keys_can_be_None=False,
        dkeys=_DMAT_KEYS,
    )

    # -----------------------------------
    # check energy / transmission values

    dmat['energy'], dmat['transmission'] = _dmat_energy_trans(
        energ=dmat['energy'],
        trans=dmat['transmission'],
    )

    # ----------
    # dref

    kne = f'{key}-nE'
    ne = dmat['energy'].size
    dref = {
        kne: {'size': ne},
    }

    # ----------
    # ddata

    kE = f'{key}-E'
    ktrans = f'{key}-trans'

    ddata = {
        kE: {
            'data': dmat['energy'],
            'ref': kne,
            'dim': 'energy',
            'quant': 'energy',
            'name': 'E',
            'units': 'eV',
        },
        ktrans: {
            'data': dmat['transmission'],
            'ref': kne,
            'dim': None,
            'quant': 'trans. coef.',
            'name': '',
            'units': '',
        },
    }

    # -----------
    # dmat

    dmat['energy'] = kE
    dmat['transmission'] = ktrans

    return dref, ddata, dmat



# coding utf-8

# Built-in
import itertools as itt
import warnings

# Common
import numpy as np
# import scipy.integrate as scpinteg
import matplotlib.pyplot as plt
from matplotlib import gridspec
import matplotlib as mpl
import matplotlib.transforms as transforms
import matplotlib.lines as mlines
import matplotlib.colors as mcolors
# from mpl_toolkits.axes_grid1 import make_axes_locatable


# tofu
from tofu.version import __version__
from .. import utils as utils
from . import _def as _def
from . import _generic_check
from . import _DataCollection_plot_text


__all__ = [
    'plot_TimeTraceColl',
]


# __author_email__ = 'didier.vezinet@cea.fr'
__github = 'https://github.com/ToFuProject/tofu/issues'
_WINTIT = 'tofu-%s        report issues / requests at {}'.format(
    __version__, __github,
)
_nchMax, _ntMax, _nfMax, _nlbdMax = 4, 3, 3, 3
_fontsize = 8
_labelpad = 0
_lls = ['-', '--', '-.', ':']
_lct = [plt.cm.tab20.colors[ii] for ii in [0, 2, 4, 1, 3, 5]]
_lcch = [plt.cm.tab20.colors[ii] for ii in [6, 8, 10, 7, 9, 11]]
_lclbd = [plt.cm.tab20.colors[ii] for ii in [12, 16, 18, 13, 17, 19]]
_lcm = _lclbd
_cbck = (0.8, 0.8, 0.8)
_dmarker = {'ax': 'o', 'x': 'x'}


_OVERHEAD = True
_CROSS = True
_DRAW = True
_CONNECT = True
_AXGRID = False
_LIB = 'mpl'
_BCKCOLOR = 'w'

_LCOLOR_DICT = [
    [
        'tab:blue', 'tab:orange', 'tab:green', 'tab:red', 'tab:purple',
        'tab:brown', 'tab:pink', 'tab:gray', 'tab:olive', 'tab:cyan',
    ],
    ['r', 'g', 'b'],
    ['m', 'y', 'c'],
]


#############################################
#############################################
#       TimeTraceCollection plots
#############################################
#############################################


def _check_proj(proj=None, allowed=None):
    """ Return validated proj as a list of keys """

    # Check allowed
    c0 = (
        isinstance(allowed, list)
        and all([isinstance(ss, str) for ss in allowed])
    )
    if not c0:
        msg = (
            "Arg allowed must be list of str!\n"
            + "\t- provided: {}".format(allowed)
        )
        raise Exception(msg)

    # Check proj
    if proj is None:
        proj = 'all'
    if proj == 'all':
        proj = allowed
    if isinstance(proj, str):
        proj = [proj]
    c0 = (
        isinstance(proj, list)
        and all([isinstance(ss, str) and ss in allowed for ss in proj])
    )
    if not c0:
        msg = (
            "Arg proj must be a list of str from {}\n".format(allowed)
            + "\t- provided: {}".format(proj)
        )
        raise Exception(msg)

    return proj


#############################################
#############################################
#       TimeTraceCollection plots
#############################################
#############################################


def _get_fig_dax_mpl(dcases=None, axgrid=None,
                     overhead=False, novert=1, noverch=1,
                     ntmax=None, nchmax=None,
                     sharex_t=None, sharey_t=None,
                     sharex_ch=None, sharey_ch=None,
                     cross=False, share_cross=None, cross_unique=None,
                     bckcolor=None, wintit=None, tit=None):

    # -----------------
    # Check format input
    # -----------------

    assert isinstance(nax, int)
    lc = [axgrid is None, axgrid is True, isinstance(axgrid, tuple)]
    assert any(lc)

    if bckcolor is None:
        bckcolor = _BCKCOLOR
    if wintit is None:
        wintit = _WINTIT

    if axgrid is None:
        axgrid = _AXGRID
    if axgrid is None:
        axgrid = False
    if axgrid is not False:
        assert dim == 1
    if dmargin is None:
        dmargin = _def.dmargin1D

    # -----------------
    # Check all cases
    # -----------------

    if axgrid is not False:
        # Only time traces
        assert all([vv['dim'] == 1 for vv in dcases.values()])
        ncases = len(dcases)
    elif any([vv.get('isspectral', False) for vv in dcases.values()]):
        # Spectral case => only one case !
        assert len(dcases) == i
        isspectral = True
    else:
        # All time traces to overhead, not counted in ncases
        ncases = 0
        for kk, vv in dcases.items():
            if vv['dim'] > 1:
                if vv['is2d']:
                    dcases[kk]['naxch'] = nchmax
                else:
                    dcases[kk]['naxch'] = 1
                ncases += 1
        isspectral = False

    # Options
    # (A) time traces vignettes
    # (B) time traces with or w/o cross + overhead
    # (C) profiles1d / cam1d with or w/o cross + overhead
    # (D) cam2d with or w/o cross + overhead, nt = 1, 2
    # (E) cam1dspectral with or w/o cross + overhead, nch = 1, 2
    # (F) profile2d with or w/o cross + overhead, nch = 1, 2

    # -----------------
    # Make figure
    # -----------------

    fs = utils.get_figuresize(fs)
    fig = plt.figure(facecolor=bckcolor, figsize=fs)
    if wintit is not False:
        fig.canvas.manager.set_window_title(wintit)

    # -----------------
    # Check all cases
    # -----------------

    dax = {'lkey': 0, 'dict': {}}
    if axgrid is False:
        if isspectral:
            naxvgrid = nchMax + 1 + overhead
        else:
            naxvgrid = ncases + overhead
        naxhgrid = 2*2 + cross
        gridax = gridspec.GridSpec(2*naxvgrid, naxhgrid, **dmargin)

        if cD and ntMax == 2:
            naxh = naxhgrid + 1

        # Create overead t and ch
        if overhead:
            for ii in range(novert):
                key = 'over-t{}'.format(ii)
                i0, i1 = ii*novert, (ii+1)*novert
                dax['dict'][key] = fig.add_subplot(gridax[i0:i1, :2])
            for ii in range(noverch):
                key = 'over-ch{}'.format(ii)
                i0, i1 = ii*noverch, (ii+1)*noverch
                dax['dict'][key] = fig.add_subplot(gridax[i0:i1, 2:4])

            # Add hor
            if cross:
                dax['dict']['hor'] = fig.add_subplot(gridax[:2, 4:])

        # Create cross
        i0 = 2*overhead
        if cross:
            if cross_unique:
                dax['dict']['cross'] = fig.add_subplot(gridax[i0:, 4:])
            else:
                for ii in range(ncases):
                    key = 'cross{}'.format(ii)
                    ii0 = i0+ii*2
                    dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, 4:])

        # Create time and channel axes
        for ii in range(ncases):
            key = 't{}'.format(ii)
            ii0 = i0+ii*2
            dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, :2])

        dax['ch'] = []
        for ii in range(ncases):
            if dcases[lcases[ii]].get('is2D', False):
                ii0 = i0+ii*2
                for jj in range(nchmax):
                    key = 'ch{}-{}'.format(ii, jj)
                    dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, 2+jj])
            else:
                for ii in range(ncases):
                    key = 'ch{}'.format(ii)
                    ii0 = i0+ii*2
                    dax['dict'][key] = fig.add_subplot(gridax[ii0:ii0+2, 2:4])

    else:
        if axgrid is True:
            nax = int(np.ceil(np.sqrt(ncases)))
            naxvgrid = nax
            naxhgrid = int(np.ceil(ncases / nax))
        else:
            naxvgrid, naxhgrid = axgrid
        assert naxvgrid*naxhgrid >= ncases
        axgrid = gridspec.GridSpec(naxvgrid, naxhgrid+cross, **dmargin)
        for ii in range(ncase):
            i0 = ii % naxvgrid
            i1 = ii - i0*naxhgrid
            key = 't{}-{}'.format(i0, i1)
            dax['dict'][key] = fig.add_subplot(gridax[i0, i1])

        if cross:
            dax['dict']['cross'] = fig.add_subplot(gridax[:, -1])

    dax['lkey'] = sorted(list(dax['dict'].keys()))
    dax['fig'] = fig
    dax['can'] = fig.canvas
    return dax


def plot_DataColl(coll, overhead=None,
                  color=None, ls=None, marker=None, ax=None,
                  cross=None, share_cross=None, cross_unique=None,
                  axgrid=None, dmargin=None, legend=None,
                  fs=None, draw=None, connect=None, lib=None):

    # --------------------
    # Check / format input
    # --------------------

    if overhead is None:
        overhead = _OVERHEAD
    if cross is None:
        if 'Plasma' in coll.__class__.__name__:
            cross = _CROSS
        else:
            cross = False
    if share_cross is None and cross:
        share_cross = False
    if draw is None:
        draw = _DRAW
    if connect is None:
        connect = _CONNECT
    if lib is None:
        lib = _LIB

    assert lib == 'mpl', 'Only matplotlib available so far !'

    # --------------------
    # Get keys of data to plot
    # --------------------

    if len(coll.dgroup) == 1:
        # TimeTraces
        dcases = None
    else:
        pass

    # --------------------
    # Get graphics dict of keys
    # --------------------

    # Case with time traces only
    daxg, lparam = {}, coll.lparam
    laxt = [('ax', ax), ('color', color), ('ls', ls), ('marker', marker)]
    for ss, vv in laxt:
        if vv is None:
            daxg[ss] = None
            continue

        if vv in lparam:

            # get keys with matching vv
            lp = coll.get_param(vv, key=lk, returnas=str)
            dv = {pp: [kk for kk in lk if self._ddata['dict'][kk][pp] == pp]
                  for pp in set(lp)}

            # Set new param ss
            if ss not in lparam:
                coll.add_param(ss, value=None)      # TBF

        daxg[ss] = dv

    # Case with any type of data => only valid for time traces (True ?)

    # Get number of axes

    # --------------------
    # Prepare figure / axes
    # --------------------

    # Get array of axes positions as a dict
    dim = len(coll.lgroup)
    config = None
    spectral = coll.isspectral

    if lib == 'mpl':
        daxg = _get_fig_daxg_mpl(dcases=dcases, axgrid=axgrid,
                                 cross=cross, overhead=overhead,
                                 ntmax=ntmax, nchmax=nchmax)

    # --------------------
    # Populate axes with static
    # --------------------

    # --------------------
    # Populate axes with dynamic (dobj)
    # --------------------
    dobj = {}

    # --------------------
    # Interactivity
    # --------------------
    collplot = None

    return collplot



# -*- coding: utf-8 -*-


# Built-in


# Common
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.colors as mcolors
import datastock as ds

# specific
from . import _generic_check
from . import _class1_compute as _compute


# #############################################################################
# #############################################################################
#                           plot mesh
# #############################################################################


def _plot_mesh_check(
    coll=None,
    key=None,
    ind_knot=None,
    ind_cent=None,
    crop=None,
    bck=None,
    color=None,
    dleg=None,
):

    # key
    key = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=list(coll.dobj.get('mesh', {}).keys()),
    )

    # crop, bck
    crop = ds._generic_check._check_var(crop, 'crop', default=True, types=bool)
    bck = ds._generic_check._check_var(bck, 'bck', default=True, types=bool)

    # ind_knot
    if ind_knot is not None:
        ind_knot = coll.select_mesh_elements(
            key=key, ind=ind_knot, elements='knots',
            returnas='data', return_neighbours=True, crop=crop,
        )

    # ind_cent
    if ind_cent is not None:
        ind_cent = coll.select_mesh_elements(
            key=key, ind=ind_cent, elements='cents',
            returnas='data', return_neighbours=True, crop=crop,
        )

    # color
    if color is None:
        color = 'k'
    if not mcolors.is_color_like(color):
        msg = (
            "Arg color must be a valid matplotlib color identifier!\n"
            f"Provided: {color}"
        )
        raise Exception(msg)

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    return key, ind_knot, ind_cent, crop, bck, color, dleg


def _plot_mesh_prepare(
    coll=None,
    key=None,
    crop=None,
    bck=None,
):

    # --------
    # prepare

    meshtype = coll.dobj['mesh'][key]['type']

    grid_bck = None
    if meshtype == 'rect':
        Rk, Zk = coll.dobj['mesh'][key]['knots']
        R = coll.ddata[Rk]['data']
        Z = coll.ddata[Zk]['data']

        vert = np.array([
            np.repeat(R, 3),
            np.tile((Z[0], Z[-1], np.nan), R.size),
        ])
        hor = np.array([
            np.tile((R[0], R[-1], np.nan), Z.size),
            np.repeat(Z, 3),
        ])

        # --------
        # compute

        if crop is False or coll.dobj['mesh'][key]['crop'] is False:
            grid = np.concatenate((vert, hor), axis=1)

        else:

            crop = coll.ddata[coll.dobj['mesh'][key]['crop']]['data']

            grid = []
            icropR = np.r_[range(R.size-1), R.size-2]
            jcropZ = np.r_[range(Z.size-1), Z.size-2]

            # vertical lines  TBC
            for ii, ic in enumerate(icropR):
                if np.any(crop[ic, :]):
                    if ii in [0, R.size-1]:
                        cropi = crop[ic, :]
                    else:
                        cropi = crop[ic, :] | crop[ic-1, :]
                    lseg = []
                    for jj, jc in enumerate(jcropZ):
                        if jj == 0 and cropi[jc]:
                            lseg.append(Z[jj])
                        elif jj == Z.size-1 and cropi[jc]:
                            lseg.append(Z[jj])
                        elif cropi[jc] and not cropi[jc-1]:
                            if len(lseg) > 0:
                                lseg.append(np.nan)
                            lseg.append(Z[jj])
                        elif (not cropi[jc]) and cropi[jc-1]:
                            lseg.append(Z[jc])
                    grid.append(np.concatenate(
                        (
                            np.array([R[ii]*np.ones((len(lseg),)), lseg]),
                            np.full((2, 1), np.nan)
                        ),
                        axis=1,
                    ))

            # horizontal lines
            for jj, jc in enumerate(jcropZ):
                if np.any(crop[:, jc]):
                    if jj in [0, Z.size-1]:
                        cropj = crop[:, jc]
                    else:
                        cropj = crop[:, jc] | crop[:, jc-1]
                    lseg = []
                    for ii, ic in enumerate(icropR):
                        if ii in [0, R.size-1] and cropj[ic]:
                            lseg.append(R[ii])
                        elif cropj[ic] and not cropj[ic-1]:
                            if len(lseg) > 0:
                                lseg.append(np.nan)
                            lseg.append(R[ii])
                        elif (not cropj[ic]) and cropj[ic-1]:
                            lseg.append(R[ic])
                    grid.append(np.concatenate(
                        (
                            np.array([lseg, Z[jj]*np.ones((len(lseg),))]),
                            np.full((2, 1), np.nan)
                        ),
                        axis=1,
                    ))

            grid = np.concatenate(tuple(grid), axis=1)

            if bck is True:
                grid_bck = np.concatenate((vert, hor), axis=1)

    else:
        kknots = coll.dobj['mesh'][key]['knots']
        R = coll.ddata[kknots[0]]['data']
        Z = coll.ddata[kknots[1]]['data']

        indtri = coll.ddata[coll.dobj['mesh'][key]['ind']]['data']

        # find unique segments from all triangles
        segs = np.unique(
            np.sort(np.concatenate(
                (indtri[:, 0:2], indtri[:, 1:], indtri[:, ::2]),
                axis=0,
            )),
            axis=0,
        )

        # build long segments if possible
        ind = np.ones((segs.shape[0],), dtype=bool)
        ind[0] = False
        lseg = [segs[0, :]]
        last = segs[0, :]
        while np.any(ind):
            ii = segs[ind, 0] == last[-1]
            if np.any(ii):
                ii = ind.nonzero()[0][ii]
                dR0 = R[last[1]] - R[last[0]]
                dZ0 = Z[last[1]] - Z[last[0]]
                dR = np.diff(R[segs[ii, :]], axis=1)[:, 0]
                dZ = np.diff(Z[segs[ii, :]], axis=1)[:, 0]
                norm0 = np.sqrt(dR0**2 + dZ0**2)
                norm = np.sqrt(dR**2 + dZ**2)
                sca = (dR0*dR + dZ0*dZ) / (norm0 * norm)
                iwin = ii[np.argmax(sca)]
                lseg.append([segs[iwin, 1]])

            else:
                lseg.append([-1])
                iwin = ind.nonzero()[0][0]
                lseg.append(segs[iwin, :])

            last = segs[iwin, :]
            ind[iwin] = False

        lseg = np.concatenate(lseg)
        grid = np.array([R[lseg], Z[lseg]])
        grid[0, lseg == -1] = np.nan

    return grid, grid_bck


def _plot_mesh_prepare_polar_cont(
    coll=None,
    key=None,
    k2d=None,
    RR=None,
    ZZ=None,
    ind=None,
    nn=None,
):

    # ---------------------
    # sample mesh if needed

    # ---------------------
    # get map of rr / angle

    if callable(k2d):

        # check RR
        if RR is None:
            msg = (
                "radius2d / angle2d are callable => provide RR and ZZ!"
            )
            raise Exception(msg)

        # compute map
        rr = k2d(RR, ZZ)[None, ...]
        assert rr.ndim == RR.ndim + 1
        reft = None
        nt = 1

        if nn is None:
            nn = 50

        # create vector
        rad = np.linspace(np.nanmin(rr), np.nanmax(rr), nn)

    else:
        kn = coll.dobj[coll._which_mesh][key]['knots'][ind]
        rad = coll.ddata[kn]['data']
        kb2 = coll.ddata[k2d]['bsplines']

        if RR is None:
            km2 = coll.dobj['bsplines'][kb2]['mesh']
            RR, ZZ = coll.get_sample_mesh(
                key=km2,
                res=None,
                grid=True,
                mode=None,
                R=None,
                Z=None,
                DR=None,
                DZ=None,
                imshow=True,
            )

        rr = coll.interpolate_profile2d(
            key=k2d,
            R=RR,
            Z=ZZ,
            grid=False,
            return_params=False,
        )[0]

        refr2d = coll.ddata[k2d]['ref']
        refbs = coll.dobj['bsplines'][kb2]['ref']
        if refr2d == refbs:
            reft = None
            nt = 1
            rr = rr[None, ...]
        elif len(refr2d) == len(refbs) + 1 and refr2d[1:] == refbs:
            reft = refr2d[0]
            nt = coll.dref[reft]['size']

    assert rr.shape[0] == nt

    # ----------------
    # Compute contours

    contR, contZ = _compute._get_contours(
        RR=RR,
        ZZ=ZZ,
        val=rr,
        levels=rad,
    )

    # refrad
    refrad = coll.dobj[coll._which_mesh][key]['knots'][ind]

    return contR, contZ, rad, reft, refrad, RR, ZZ


def _plot_mesh_prepare_polar(
    coll=None,
    key=None,
    # Necessary for callable radius2d
    RR=None,
    ZZ=None,
):

    # --------
    # prepare

    # create rectangular grid and compute radius at each point
    k2d = coll.dobj[coll._which_mesh][key]['radius2d']
    (
        contRrad, contZrad,
        rad, reft, refrad,
        RR, ZZ,
    ) = _plot_mesh_prepare_polar_cont(
        coll=coll,
        key=key,
        k2d=k2d,
        RR=RR,
        ZZ=ZZ,
        ind=0,
        nn=None,        # nrad if k2d callable
    )

    # -----------
    # contour of angle if angle not None

    contRang, contZang, ang, refang = None, None, None, None
    if len(coll.dobj[coll._which_mesh][key]['shape-c']) == 2:
        # create rectangular grid and compute radius at each point
        k2d = coll.dobj[coll._which_mesh][key]['angle2d']
        (
            contRang, contZang,
            ang, _, refang,
            _, _,
        ) = _plot_mesh_prepare_polar_cont(
            coll=coll,
            key=key,
            k2d=k2d,
            RR=RR,
            ZZ=ZZ,
            ind=1,
            nn=None,        # nang if k2d callable
        )

    return (
        contRrad, contZrad, rad, refrad,
        contRang, contZang, ang, refang,
        reft,
    )


def plot_mesh(
    coll=None,
    key=None,
    ind_knot=None,
    ind_cent=None,
    crop=None,
    bck=None,
    nmax=None,
    color=None,
    dax=None,
    dmargin=None,
    fs=None,
    dleg=None,
    connect=None,
):
    """ Plot the desired mesh

    rect and tri meshes are constant
    polar meshes can vary in time

    """

    # --------------
    # check input

    key, ind_knot, ind_cent, crop, bck, color, dleg = _plot_mesh_check(
        coll=coll,
        key=key,
        ind_knot=ind_knot,
        ind_cent=ind_cent,
        crop=crop,
        bck=bck,
        color=color,
        dleg=dleg,
    )

    # ------------------------
    # call appropriate routine

    if coll.dobj[coll._which_mesh][key]['type'] in ['rect', 'tri']:
        # time-fixed meshes
        return _plot_mesh_recttri(
            coll=coll,
            key=key,
            ind_knot=ind_knot,
            ind_cent=ind_cent,
            crop=crop,
            bck=bck,
            color=color,
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            dleg=dleg,
        )

    else:
        # possibly time-varying mesh
        return _plot_mesh_polar(
            coll=coll,
            key=key,
            nmax=nmax,
            color=color,
            dax=dax,
            fs=fs,
            dmargin=dmargin,
            dleg=dleg,
            connect=connect,
        )


def _plot_mesh_recttri(
    coll=None,
    key=None,
    ind_knot=None,
    ind_cent=None,
    crop=None,
    bck=None,
    color=None,
    dax=None,
    fs=None,
    dmargin=None,
    dleg=None,
):

    # --------------
    #  Prepare data

    grid, grid_bck = _plot_mesh_prepare(
        coll=coll,
        key=key,
        crop=crop,
        bck=bck,
    )

    # --------------
    # plot - prepare

    if dax is None:

        if dmargin is None:
            dmargin = {
                'left': 0.1, 'right': 0.9,
                'bottom': 0.1, 'top': 0.9,
                'hspace': 0.1, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0], aspect='equal')
        ax0.set_xlabel(f'R (m)')
        ax0.set_ylabel(f'Z (m)')

        dax = {'cross': ax0}

    dax = _generic_check._check_dax(dax=dax, main='cross')

    # --------------
    # plot

    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        if grid_bck is not None and bck is True:
            ax.plot(
                grid_bck[0, :],
                grid_bck[1, :],
                ls='-',
                lw=0.5,
                color=color,
                alpha=0.5,
                label=key,
            )

        ax.plot(
            grid[0, :],
            grid[1, :],
            color=color,
            ls='-',
            lw=1.,
            label=key,
        )

        if ind_knot is not None:
            ax.plot(
                ind_knot[0][0],
                ind_knot[0][1],
                marker='o',
                ms=8,
                ls='None',
                color=color,
                label='knots',
            )
            ax.plot(
                ind_knot[1][0, :, :],
                ind_knot[1][1, :, :],
                marker='x',
                ms=4,
                ls='None',
                color=color,
            )

        if ind_cent is not None:
            ax.plot(
                ind_cent[0][0],
                ind_cent[0][1],
                marker='x',
                ms=8,
                ls='None',
                color=color,
                label='cents',
            )
            ax.plot(
                ind_cent[1][0, :, :],
                ind_cent[1][1, :, :],
                marker='o',
                ms=4,
                ls='None',
                color=color,
            )

    # --------------
    # dleg

    if dleg is not False:
        for kax in dax.keys():
            dax[kax]['handle'].legend(**dleg)

    return dax


def _plot_mesh_polar(
    coll=None,
    key=None,
    npts=None,
    nmax=None,
    color=None,
    dax=None,
    fs=None,
    dmargin=None,
    dleg=None,
    connect=None,
):

    # --------------
    #  Prepare data

    if nmax is None:
        nmax = 2

    (
        contRrad, contZrad, rad, refrad,
        contRang, contZang, ang, refang,
        reft,
    ) = _plot_mesh_prepare_polar(
        coll=coll,
        key=key,
    )
    refptsr = 'ptsr'
    nt, nr, nptsr = contRrad.shape
    if contRang is not None:
        refptsa = 'ptsa'
        _, nang, nptsa = contRang.shape

    # --------------------
    # Instanciate Plasma2D

    coll2 = coll.__class__()

    # ref
    coll2.add_ref(
        key=reft,
        size=nt,
    )
    reft = list(coll2.dref.keys())[0]
    coll2.add_ref(
        key=refrad,
        size=nr,
    )
    coll2.add_ref(
        key=refptsr,
        size=nptsr,
    )

    if contRang is not None:
        coll2.add_ref(
            key=refang,
            size=nang,
        )
        coll2.add_ref(
            key=refptsa,
            size=nptsa,
        )

    # data
    coll2.add_data(
        key='radius',
        data=rad,
        ref=(refrad,)
    )
    coll2.add_data(
        key='contRrad',
        data=contRrad,
        ref=(reft, refrad, refptsr)
    )
    coll2.add_data(
        key='contZrad',
        data=contZrad,
        ref=(reft, refrad, refptsr)
    )

    if contRang is not None:
        coll2.add_data(
            key='angle',
            data=ang,
            ref=(refang,)
        )
        coll2.add_data(
            key='contRang',
            data=contRang,
            ref=(reft, refang, refptsa)
        )
        coll2.add_data(
            key='contZang',
            data=contZang,
            ref=(reft, refang, refptsa)
        )

    # -----
    # plot

    if contRang is None:
        return coll2.plot_as_mobile_lines(
            keyX='contRrad',
            keyY='contZrad',
            key_time=reft,
            key_chan='radius',
            connect=connect,
        )

    else:

        daxrad, dgrouprad = coll2.plot_as_mobile_lines(
            keyX='contRrad',
            keyY='contZrad',
            key_time=reft,
            key_chan='radius',
            connect=False,
            inplace=False,
        )

        daxang, dgroupang = coll2.plot_as_mobile_lines(
            keyX='contRang',
            keyY='contZang',
            key_time=reft,
            key_chan='angle',
            connect=False,
            inplace=False,
        )

        # connect
        if connect is False:
            return (daxrad, daxang), (dgrouprad, dgroupang)

        else:
            daxrad.setup_interactivity(
                kinter='inter0', dgroup=dgrouprad, dinc=None,
            )
            daxrad.disconnect_old()
            daxrad.connect()

            daxang.setup_interactivity(
                kinter='inter0', dgroup=dgroupang, dinc=None,
            )
            daxang.disconnect_old()
            daxang.connect()

            daxrad.show_commands()
            return daxrad, daxang


# #############################################################################
# #############################################################################
#                           plot bspline
# #############################################################################


def _plot_bsplines_get_dRdZ(coll=None, km=None, meshtype=None):
    # Get minimum distances

    if meshtype == 'rect':
        kR, kZ = coll.dobj['mesh'][km]['knots']
        Rk = coll.ddata[kR]['data']
        Zk = coll.ddata[kZ]['data']
        dR = np.min(np.diff(Rk))
        dZ = np.min(np.diff(Zk))

    elif meshtype == 'tri':
        indtri = coll.ddata[coll.dobj['mesh'][km]['ind']]['data']
        kknots = coll.dobj['mesh'][km]['knots']
        Rk = coll.ddata[kknots[0]]['data']
        Zk = coll.ddata[kknots[1]]['data']
        R = Rk[indtri]
        Z = Zk[indtri]
        dist = np.mean(np.array([
            np.sqrt((R[:, 1] - R[:, 0])**2 + (Z[:, 1] - Z[:, 0])**2),
            np.sqrt((R[:, 2] - R[:, 1])**2 + (Z[:, 2] - Z[:, 1])**2),
            np.sqrt((R[:, 2] - R[:, 0])**2 + (Z[:, 2] - Z[:, 0])**2),
        ]))
        dR, dZ = dist, dist

    else:
        km2 = coll.dobj[coll._which_mesh][km]['submesh']
        meshtype = coll.dobj[coll._which_mesh][km2]['type']
        return _plot_bsplines_get_dRdZ(
            coll=coll, km=km2, meshtype=meshtype,
        )

    Rminmax = [Rk.min(), Rk.max()]
    Zminmax = [Zk.min(), Zk.max()]
    return dR, dZ, Rminmax, Zminmax


def _plot_bspline_check(
    coll=None,
    key=None,
    indbs=None,
    indt=None,
    knots=None,
    cents=None,
    plot_mesh=None,
    cmap=None,
    dleg=None,
):

    # key
    lk = list(coll.dobj.get('bsplines', {}).keys())
    key = ds._generic_check._check_var(
        key, 'key',
        default=None,
        types=str,
        allowed=lk,
    )
    keym0 = coll.dobj['bsplines'][key]['mesh']
    mtype0 = coll.dobj[coll._which_mesh][keym0]['type']
    if mtype0 == 'polar':
        keym = coll.dobj[coll._which_mesh][keym0]['submesh']
        mtype = coll.dobj[coll._which_mesh][keym]['type']
    else:
        keym = keym0
        mtype = mtype0

    # knots, cents
    knots = ds._generic_check._check_var(
        knots, 'knots', default=True, types=bool,
    )
    cents = ds._generic_check._check_var(
        cents, 'cents', default=True, types=bool,
    )

    # ind_bspline
    if indbs is not None:
        indbs = coll.select_bsplines(
            key=key,
            ind=indbs,
            returnas='ind',
            return_knots=False,
            return_cents=False,
            crop=False,
        )

    _, knotsi, centsi = coll.select_bsplines(
        key=key,
        ind=indbs,
        returnas='data',
        return_knots=True,
        return_cents=True,
        crop=False,
    )

    # indt
    nt = False
    if mtype0 == 'polar':
        radius2d = coll.dobj[coll._which_mesh][keym0]['radius2d']
        r2d_reft = coll.get_time(key=radius2d)[2]
        if r2d_reft is not None:
            nt = coll.dref[r2d_reft]['size']

    if nt is False:
        indt = None
    else:
        if indt is None:
            indt = 0
        indt = np.atleast_1d(indt).ravel()[0]

    # plot_mesh
    plot_mesh = ds._generic_check._check_var(
        plot_mesh, 'plot_mesh',
        default=True,
        types=bool,
    )

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    return (
        key, keym0, keym, mtype0, mtype,
        indbs, indt,
        knots, cents, knotsi, centsi,
        plot_mesh, cmap, dleg,
    )


def _plot_bspline_prepare(
    coll=None,
    # keys
    key=None,
    keym0=None,
    keym=None,
    mtype0=None,
    mtype=None,
    # indices
    indbs=None,
    indt=None,
    # options
    res=None,
    knotsi=None,
    centsi=None,
    val_out=None,
    nan0=None,
):

    # check input
    deg = coll.dobj['bsplines'][key]['deg']

    # get dR, dZ
    dR, dZ, _, _ = _plot_bsplines_get_dRdZ(
        coll=coll, km=keym, meshtype=mtype,
    )

    # resolution of sampling
    if res is None:
        if mtype == 'rect':
            res_coef = 0.05
        else:
            res_coef = 0.25
        res = [res_coef*dR, res_coef*dZ]

    # sampling domain
    if mtype0 == 'polar':
        DR = None
        DZ = None
    else:
        knotsiR, knotsiZ = knotsi
        DR = [np.nanmin(knotsiR) + dR*1.e-10, np.nanmax(knotsiR) - dR*1.e-10]
        DZ = [np.nanmin(knotsiZ) + dZ*1.e-10, np.nanmax(knotsiZ) - dZ*1.e-10]

    # sample
    R, Z = coll.get_sample_mesh(
        key=keym,
        res=res,
        DR=DR,
        DZ=DZ,
        mode='abs', grid=True, imshow=True,
    )

    # bspline
    bspline = coll.interpolate_profile2d(
        key=key,
        R=R,
        Z=Z,
        # coefs=coefs,
        indt=indt,
        indbs=indbs,
        details=indbs is not None,
        grid=False,
        nan0=nan0,
        val_out=val_out,
        return_params=False,
    )[0]

    if indbs is None:
        if bspline.ndim == R.ndim + 1:
            assert bspline.shape[1:] == R.shape
            bspline = bspline[0, ...]
    else:
        if bspline.ndim == R.ndim + 1:
            assert bspline.shape[:-1] == R.shape
            bspline = np.nansum(bspline, axis=-1)
        elif bspline.ndim == R.ndim + 2:
            assert bspline.shape[1:-1] == R.shape
            bspline = np.nansum(bspline[0, ...], axis=-1)

    if bspline.shape != R.shape:
        import pdb; pdb.set_trace() # DB
        pass

    # extent
    if mtype0 == 'polar':
        extent = (
            R.min(), R.max(),
            Z.min(), Z.max(),
        )
    else:
        extent = (
            DR[0], DR[1],
            DZ[0], DZ[1],
        )

    # interpolation
    if deg == 0:
        interp = 'nearest'
    elif deg == 1:
        interp = 'bilinear'
    elif deg >= 2:
        interp = 'bicubic'

    return bspline, extent, interp


def plot_bspline(
    # ressources
    coll=None,
    # inputs
    key=None,
    indbs=None,
    indt=None,
    # parameters
    knots=None,
    cents=None,
    res=None,
    plot_mesh=None,
    val_out=None,
    nan0=None,
    # plot-specific
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dleg=None,
):

    # --------------
    # check input

    (
        key, keym0, keym, mtype0, mtype,
        indbs, indt,
        knots, cents, knotsi, centsi,
        plot_mesh, cmap, dleg,
    ) = _plot_bspline_check(
        coll=coll,
        key=key,
        indbs=indbs,
        indt=indt,
        knots=knots,
        cents=cents,
        plot_mesh=plot_mesh,
        cmap=cmap,
        dleg=dleg,
    )

    # --------------
    #  Prepare data

    bspline, extent, interp = _plot_bspline_prepare(
        coll=coll,
        key=key,
        keym=keym,
        mtype0=mtype0,
        mtype=mtype,
        indbs=indbs,
        indt=indt,
        knotsi=knotsi,
        centsi=centsi,
        res=res,
        val_out=val_out,
        nan0=nan0,
    )

    # --------------
    # plot - prepare

    if dax is None:

        if dmargin is None:
            dmargin = {
                'left': 0.1, 'right': 0.9,
                'bottom': 0.1, 'top': 0.9,
                'hspace': 0.1, 'wspace': 0.1,
            }

        fig = plt.figure(figsize=fs)
        gs = gridspec.GridSpec(ncols=1, nrows=1, **dmargin)
        ax0 = fig.add_subplot(gs[0, 0], aspect='equal')
        ax0.set_xlabel(f'R (m)')
        ax0.set_ylabel(f'Z (m)')

        dax = {'cross': ax0}

    dax = _generic_check._check_dax(dax=dax, main='cross')

    # --------------
    # plot

    if plot_mesh is True:
        keym = coll.dobj['bsplines'][key]['mesh']
        if mtype0 == 'polar':
            _ = coll.plot_mesh(key=keym, dleg=False)
        else:
            dax = coll.plot_mesh(key=keym, dax=dax, dleg=False)

    kax = 'cross'
    if dax.get(kax) is not None:
        ax = dax[kax]['handle']

        ax.imshow(
            bspline,
            extent=extent,
            interpolation=interp,
            origin='lower',
            aspect='equal',
            cmap=cmap,
            vmin=0.,
            vmax=1.,
        )

        if mtype0 != 'polar':
            if knots is not False:
                ax.plot(
                    knotsi[0].ravel(),
                    knotsi[1].ravel(),
                    marker='x',
                    ms=6,
                    ls='None',
                    color='k',
                )

            if cents is not False:
                ax.plot(
                    centsi[0].ravel(),
                    centsi[1].ravel(),
                    marker='o',
                    ms=6,
                    ls='None',
                    color='k',
                )

        ax.relim()
        ax.autoscale()

        # --------------
        # dleg

        if dleg is not False:
            ax.legend(**dleg)

    return dax


# #############################################################################
# #############################################################################
#                           plot profile2d
# #############################################################################


def _plot_profile2d_check(
    coll=None,
    key=None,
    cmap=None,
    dcolorbar=None,
    dleg=None,
):

    # key
    dk = coll.get_profiles2d()
    key = ds._generic_check._check_var(
        key, 'key', types=str, allowed=list(dk.keys())
    )
    keybs = dk[key]
    refbs = coll.dobj['bsplines'][keybs]['ref']
    keym = coll.dobj['bsplines'][keybs]['mesh']
    mtype = coll.dobj[coll._which_mesh][keym]['type']

    # cmap
    if cmap is None:
        cmap = 'viridis'

    # dcolorbar
    defdcolorbar = {
        # 'location': 'right',
        'fraction': 0.15,
        'orientation': 'vertical',
    }
    dcolorbar = ds._generic_check._check_var(
        dcolorbar, 'dcolorbar',
        default=defdcolorbar,
        types=dict,
    )

    # dleg
    defdleg = {
        'bbox_to_anchor': (1.1, 1.),
        'loc': 'upper left',
        'frameon': True,
    }
    dleg = ds._generic_check._check_var(
        dleg, 'dleg',
        default=defdleg,
        types=(bool, dict),
    )

    # polar1d
    polar1d = False
    if mtype  == 'polar':
        if coll.dobj['bsplines'][keybs]['class'].knotsa is None:
            polar1d = True
        elif len(coll.dobj['bsplines'][keybs]['ref']) == 2:
            polar1d = True
        else:
            apbs = coll.dobj['bsplines'][keybs]['class'].apex_per_bs_a
            if np.sum([aa is not None for aa in apbs]) == 1:
                polar1d = True

    return key, keybs, keym, cmap, dcolorbar, dleg, polar1d


def _plot_profiles2d_prepare(
    coll=None,
    key=None,
    keybs=None,
    keym=None,
    coefs=None,
    indt=None,
    res=None,
    mtype=None,
):

    # check input
    deg = coll.dobj['bsplines'][keybs]['deg']

    # get dR, dZ
    dR, dZ, Rminmax, Zminmax = _plot_bsplines_get_dRdZ(
        coll=coll, km=keym, meshtype=mtype,
    )

    if res is None:
        res_coef = 0.2
        res = [res_coef*dR, res_coef*dZ]

    # compute
    coll2 = coll.interpolate_profile2d(
        key=key,
        R=None,
        Z=None,
        coefs=coefs,
        indt=indt,
        res=res,
        details=False,
        nan0=True,
        imshow=False,
        return_params=False,
        store=True,
        inplace=False,
    )

    keymap = [k0 for k0, v0 in coll2.ddata.items() if v0['data'].ndim > 1][0]
    ndim = coll2.ddata[keymap]['data'].ndim

    refmap = coll2.ddata[keymap]['ref']
    dkeys = {
        'key': keymap,
        'keyX': coll2.get_ref_vector(key=keymap, ref=refmap[-2])[3],
        'keyY': coll2.get_ref_vector(key=keymap, ref=refmap[-1])[3],
        'keyZ': None,
    }
    if ndim == 3:
        keyZ = coll2.get_ref_vector(key=keymap, ref=refmap[0])[3]
        import datastock as ds
        uniform = ds._plot_as_array._check_uniform_lin(
            k0=keyZ, ddata=coll2.ddata,
        )
        if not uniform:
            keyZ = None

    if deg == 0:
        interp = 'nearest'
    elif deg == 1:
        interp = 'bilinear'
    elif deg >= 2:
        interp = 'bicubic'

    # radial of polar
    if mtype == 'polar':
        # lcol
        lcol = ['k', 'r', 'b', 'g', 'm', 'c', 'y']

    else:
        lcol = None

    return coll2, dkeys, interp, lcol


def _plot_profile2d_polar_add_radial(
    coll=None,
    key=None,
    keym=None,
    keybs=None,
    dax=None,
):

    # key to radius
    kr2d = coll.dobj[coll._which_mesh][keym]['radius2d']
    kr = coll.dobj[coll._which_mesh][keym]['knots'][0]
    rr = coll.ddata[kr]['data']
    rad = np.linspace(rr[0], rr[-1], rr.size*20)

    # get angle if any
    clas = coll.dobj['bsplines'][keybs]['class']
    if clas.knotsa is None:
        angle = None
    elif len(clas.shapebs) == 2:
        ka = coll.dobj['bsplines'][keybs]['apex'][1]
        angle = coll.ddata[ka]['data']
    elif np.sum(clas.nbs_a_per_r > 1) == 1:
        i0 = (clas.nbs_a_per_r > 1).nonzero()[0][0]
        angle = coll.dobj['bsplines'][keybs]['class'].apex_per_bs_a[i0]
    else:
        pass

    if angle is None:
        radmap = rad
        anglemap = angle
    else:
        radmap = np.repeat(rad[:, None], angle.size, axis=1)
        anglemap = np.repeat(angle[None, :], rad.size, axis=0)

    # reft
    reft, keyt, _, dind = coll.get_time_common(keys=[key, kr2d])[1:]

    # radial total profile
    radial, t_radial, _ = coll.interpolate_profile2d(
        key=key,
        radius=radmap,
        angle=anglemap,
        grid=False,
        t=keyt,
    )

    if reft is not None and radial.ndim == radmap.ndim:
        radial = np.repeat(radial[None, ...], t_radial.size, axis=0)

    # details for purely-radial cases
    if clas.knotsa is None:
        radial_details, t_radial, _ = coll.interpolate_profile2d(
            key=keybs,
            radius=rad,
            angle=None,
            grid=False,
            details=True,
        )

        if reft is None:
            radial_details = radial_details * coll.ddata[key]['data'][None, :]
            refdet = ('nradius',)
        else:
            refdet = (reft, 'nradius')
            if reft == coll.get_time(key)[2]:
                radial_details = (
                    radial_details[None, :, :]
                    * coll.ddata[key]['data'][:, None, :]
                )
            elif key in dind.keys():
                radial_details = (
                    radial_details[None, :, :]
                    * coll.ddata[key]['data'][dind[key]['ind'], None, :]
                )

        nbs = radial_details.shape[-1]

    # add to dax
    dax.add_ref(key='nradius', size=rad.size)
    if angle is not None:
        dax.add_ref(key='nangle', size=angle.size)

    if reft is not None:
        assert radial.ndim > 1 and radial.shape[0] > 1
        # if angle is not None:
            # ref = (reft, 'nangle', 'nradius')
        # else:
        ref = (reft, 'nradius')
    else:
        # if angle is not None:
            # ref = ('nangle', 'nradius')
        # else:
        ref = 'nradius'

    # add to ddata
    kradius = 'radius'
    dax.add_data(key=kradius, data=rad, ref='nradius')
    if angle is None:
        lk = ['radial']
        dax.add_data(key=lk[0], data=radial, ref=ref)
        lkdet = [f'radial-detail-{ii}' for ii in range(nbs)]
        for ii in range(nbs):
            dax.add_data(
                key=lkdet[ii], data=radial_details[..., ii], ref=refdet,
            )

    else:
        kangle = 'angle'
        dax.add_data(key=kangle, data=angle, ref='nangle')
        lkdet = None
        lk = [f'radial-{ii}' for ii in range(angle.size)]
        if reft is None:
            for ii in range(angle.size):
                dax.add_data(key=lk[ii], data=radial[:, ii], ref=ref)
        else:
            for ii in range(angle.size):
                dax.add_data(key=lk[ii], data=radial[:, :, ii], ref=ref)

    return kradius, lk, lkdet, reft


def plot_profile2d(
    # ressources
    coll=None,
    # inputs
    key=None,
    # parameters
    coefs=None,
    indt=None,
    res=None,
    # figure
    vmin=None,
    vmax=None,
    cmap=None,
    dax=None,
    dmargin=None,
    fs=None,
    dcolorbar=None,
    dleg=None,
    # interactivity
    dinc=None,
    connect=None,
):

    # --------------
    # check input

    if connect is None:
        connect = True

    (
        key, keybs, keym, cmap, dcolorbar, dleg, polar1d,
    ) = _plot_profile2d_check(
        coll=coll,
        key=key,
        cmap=cmap,
        dcolorbar=dcolorbar,
        dleg=dleg,
    )
    mtype = coll.dobj[coll._which_mesh][keym]['type']
    hastime = coll.get_time(key=key)[0]

    # --------------
    #  Prepare data

    (
        coll2, dkeys, interp, lcol,
    ) = _plot_profiles2d_prepare(
        coll=coll,
        key=key,
        keybs=keybs,
        keym=keym,
        coefs=coefs,
        indt=indt,
        res=res,
        mtype=mtype,
    )

    # ---------------
    # call right function

    if mtype in ['rect', 'tri']:
        return coll2.plot_as_array(
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            interp=interp,
            connect=connect,
            **dkeys,
        )

    else:

        if dax is None:
            dax = _plot_profile2d_polar_create_axes(
                fs=fs,
                dmargin=dmargin,
            )

        dax, dgroup = coll2.plot_as_array(
            vmin=vmin,
            vmax=vmax,
            cmap=cmap,
            dax=dax,
            dmargin=dmargin,
            fs=fs,
            dcolorbar=dcolorbar,
            dleg=dleg,
            connect=False,
            interp=interp,
            label=True,
            **dkeys,
        )

        # ------------------
        # add radial profile to dax

        kradius, lkradial, lkdet, reft = _plot_profile2d_polar_add_radial(
            coll=coll,
            key=key,
            keym=keym,
            keybs=keybs,
            dax=dax,
        )


        assert (reft is not None) == ('Z' in dgroup.keys())
        if reft is not None and reft not in dgroup['Z']['ref']:
            dgroup['Z']['ref'].append(reft)
            dgroup['Z']['data'].append('index')

        # ------------------
        # add radial profile

        kax = 'radial'
        if dax.dax.get(kax) is not None:
            ax = dax.dax[kax]['handle']
            for ii in range(len(lkradial)):

                if reft is None:
                    l0, = ax.plot(
                        dax.ddata[kradius]['data'],
                        dax.ddata[lkradial[ii]]['data'],
                        c=lcol[ii],
                        ls='-',
                        lw=2,
                    )
                else:
                    l0, = ax.plot(
                        dax.ddata[kradius]['data'],
                        dax.ddata[lkradial[ii]]['data'][0, :],
                        c=lcol[ii],
                        ls='-',
                        lw=2,
                    )

                    kl = f"radial{ii}"
                    dax.add_mobile(
                        key=kl,
                        handle=l0,
                        refs=(reft,),
                        data=[lkradial[ii]],
                        dtype=['ydata'],
                        axes=kax,
                        ind=0,
                    )

            if lkdet is not None:
                for ii in range(len(lkdet)):
                    if reft is None:
                        l0, = ax.plot(
                            dax.ddata[kradius]['data'],
                            dax.ddata[lkdet[ii]]['data'],
                            ls='-',
                            lw=1,
                        )
                    else:
                        l0, = ax.plot(
                            dax.ddata[kradius]['data'],
                            dax.ddata[lkdet[ii]]['data'][0, :],
                            ls='-',
                            lw=1,
                        )

                        kl = f"radial_det{ii}"
                        dax.add_mobile(
                            key=kl,
                            handle=l0,
                            refs=(reft,),
                            data=[lkdet[ii]],
                            dtype=['ydata'],
                            axes=kax,
                            ind=0,
                        )

            ax.set_xlim(
                dax.ddata[kradius]['data'].min(),
                dax.ddata[kradius]['data'].max(),
            )

            if vmin is not None:
                ax.set_ylim(bottom=vmin)
            if vmax is not None:
                ax.set_ylim(top=vmax)

        # connect
        if connect is True:
            dax.setup_interactivity(kinter='inter0', dgroup=dgroup, dinc=dinc)
            dax.disconnect_old()
            dax.connect()

            dax.show_commands()
            return dax
        else:
            return dax, dgroup



def _plot_profile2d_polar_create_axes(
    fs=None,
    dmargin=None,
):

    if fs is None:
        fs = (15, 9)

    if dmargin is None:
        dmargin = {
            'left': 0.05, 'right': 0.95,
            'bottom': 0.05, 'top': 0.95,
            'hspace': 0.4, 'wspace': 0.3,
        }

    fig = plt.figure(figsize=fs)
    gs = gridspec.GridSpec(ncols=6, nrows=6, **dmargin)

    # axes for image
    ax0 = fig.add_subplot(gs[:4, 2:4], aspect='auto')

    # axes for vertical profile
    ax1 = fig.add_subplot(gs[:4, 4], sharey=ax0)

    # axes for horizontal profile
    ax2 = fig.add_subplot(gs[4:, 2:4], sharex=ax0)

    # axes for traces
    ax3 = fig.add_subplot(gs[2:4, :2])

    # axes for traces
    ax7 = fig.add_subplot(gs[:2, :2], sharey=ax2)

    # axes for text
    ax4 = fig.add_subplot(gs[:3, 5], frameon=False)
    ax5 = fig.add_subplot(gs[3:, 5], frameon=False)
    ax6 = fig.add_subplot(gs[4:, :2], frameon=False)

    # dax
    dax = {
        # data
        'matrix': {'handle': ax0, 'type': 'matrix'},
        'vertical': {'handle': ax1, 'type': 'misc'},
        'horizontal': {'handle': ax2, 'type': 'misc'},
        'traces': {'handle': ax3, 'type': 'misc'},
        'radial': {'handle': ax7, 'type': 'misc'},
        # text
        'textX': {'handle': ax4, 'type': 'text'},
        'textY': {'handle': ax5, 'type': 'text'},
        'textZ': {'handle': ax6, 'type': 'text'},
    }
    return dax



# -*- coding: utf-8 -*-


import numpy as np
import scipy.interpolate as scpinterp
import matplotlib.pyplot as plt


import Polygon as plg
from Polygon import Utils as plgUtils
import datastock as ds


from . import _class5_projections
from . import _class8_compute as _compute


# ##############################################################
# ##############################################################
#           Equivalent aperture
# ##############################################################


def equivalent_apertures(
    # resources
    coll=None,
    key=None,
    key_cam=None,
    pixel=None,
    # inital contour
    add_points=None,
    # options
    convex=None,
    harmonize=None,
    reshape=None,
    return_for_etendue=None,
    # plot
    verb=None,
    plot=None,
    store=None,
):

    # -------------
    # check inputs

    (
        key,
        key_cam,
        kref,
        cref,
        spectro,
        ispectro,
        lop_pre,
        lop_pre_cls,
        lop_post,
        lop_post_cls,
        is2d,
        shape0,
        cx,
        cy,
        cz,
        pixel,
        add_points,
        convex,
        harmonize,
        reshape,
        verb,
        plot,
        store,
    ) = _check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        pixel=pixel,
        add_points=add_points,
        convex=convex,
        harmonize=harmonize,
        reshape=reshape,
        verb=verb,
        plot=plot,
        store=store,
    )

    if pixel is None:
        pixel = np.arange(0, cx.size)

    if return_for_etendue is None:
        return_for_etendue = False

    # ---------------
    # Prepare optics

    lpoly_pre = [
        coll.get_optics_poly(
            key=k0,
            add_points=add_points,
            return_outline=False,
        )
        for k0 in lop_pre
    ]
    nop_pre = len(lop_pre)

    lout_post = [
        coll.dobj[c0][k0]['dgeom']['outline']
        for c0, k0 in zip(lop_post_cls, lop_post)
    ]
    lx0_post = [coll.ddata[pp[0]]['data'] for pp in lout_post]
    lx1_post = [coll.ddata[pp[1]]['data'] for pp in lout_post]
    nop_post = len(lop_post)

    # -------------------
    # prepare functions

    # coordinate func
    coord_x01toxyz = coll.get_optics_x01toxyz(key=kref)
    lcoord_x01toxyz_poly = [
        coll.get_optics_x01toxyz(key=oo)
        for oo in lop_post
    ]

    # pts2pts func
    if spectro:
        pts2pt = coll.get_optics_reflect_pts2pt(key=kref)
    else:
        pts2pt = None
        
    # ptsvect func
    ptsvect = coll.get_optics_reflect_ptsvect(key=kref)
    lptsvect_poly = [
        coll.get_optics_reflect_ptsvect(key=oo)
        for oo in lop_post
    ]

    # equivalent aperture func
    if spectro:
        func = _get_equivalent_aperture_spectro
    else:
        func = _get_equivalent_aperture

    # -------------------
    # prepare output

    x0 = []
    x1 = []

    # ---------------
    # loop on pixels

    # dt = np.zeros((14,), dtype=float)

    # intial polygon
    p_a = coll.get_optics_outline(key=kref, add_points=False)
    p_a = plg.Polygon(np.array([p_a[0], p_a[1]]).T)

    iok = np.ones((pixel.size,), dtype=bool)

    for ii, ij in enumerate(pixel):

        if verb is True:
            msg = f"\t- camera '{key_cam}': pixel {ii + 1} / {pixel.size}"
            end = '\n' if ii == len(pixel) - 1 else '\r'
            print(msg, end=end , flush=True)

        p0, p1 = func(
            p_a=p_a,
            pt=np.r_[cx[ij], cy[ij], cz[ij]],
            nop_pre=nop_pre,
            lpoly_pre=lpoly_pre,
            nop_post=nop_post,
            lx0_post=lx0_post,
            lx1_post=lx1_post,
            # functions
            coord_x01toxyz=coord_x01toxyz,
            lcoord_x01toxyz_poly=lcoord_x01toxyz_poly,
            pts2pt=pts2pt,
            ptsvect=ptsvect,
            lptsvect_poly=lptsvect_poly,
            # options
            add_points=add_points,
            convex=convex,
            # timing
            # dt=dt,
        )

        # convex hull
        if p0 is None or p0.size == 0:
            iok[ii] = False
        elif convex:
            p0, p1 = np.array(plgUtils.convexHull(
                plg.Polygon(np.array([p0, p1]).T)
            ).contour(0)).T

            p0, p1 = _compute._interp_poly(
                lp=[p0, p1],
                add_points=add_points,
                mode='min',
                isclosed=False,
                closed=False,
                ravel=True,
                min_threshold=1.e-5,
            )

        # append
        x0.append(p0)
        x1.append(p1)

    # --------------------
    # harmonize if necessary
    # --------------------

    if harmonize:
        ln = [p0.size if p0 is not None else 0 for p0 in x0]
        nmax = np.max(ln)
        nan = np.full((nmax,), np.nan)
        for ii in range(pixel.size):
            if x0[ii] is None:
                x0[ii] = nan
                x1[ii] = nan
            elif ln[ii] < nmax:
                ndif = nmax - ln[ii]
                irand = np.random.random(ndif)
                irand = irand + np.random.randint(0, ln[ii]-1, ndif)
                imax = np.sort(np.r_[np.arange(0, ln[ii]), irand])
                imax = np.linspace(0, ln[ii]-1, nmax)
                x0[ii] = scpinterp.interp1d(
                    np.arange(0, ln[ii]),
                    x0[ii],
                    kind='linear',
                )(imax)
                x1[ii] = scpinterp.interp1d(
                    np.arange(0, ln[ii]),
                    x1[ii],
                    kind='linear',
                )(imax)

        x0 = np.array(x0)
        x1 = np.array(x1)

    # -------------
    # xyz
    # -------------

    if return_for_etendue is True:

        pts2plane = coll.get_optics_reflect_ptsvect(
            key=kref,
            asplane=True,
        )

        px = np.full(x0.shape, np.nan)
        py = np.full(x0.shape, np.nan)
        pz = np.full(x0.shape, np.nan)
        cents0 = np.full((x0.shape[0],), np.nan)
        cents1 = np.full((x0.shape[0],), np.nan)
        area = np.full((x0.shape[0],), np.nan)

        for ii, ip in enumerate(pixel):

            if not iok[ii]:
                continue

            pxi, pyi, pzi = coord_x01toxyz(x0=x0[ii, :], x1=x1[ii, :])

            (
                px[ii, :], py[ii, :], pz[ii, :],
                _, _, _, _, _, p0, p1,
            ) = pts2plane(
                pts_x=cx[ii],
                pts_y=cy[ii],
                pts_z=cz[ii],
                vect_x=pxi - cx[ii],
                vect_y=pyi - cy[ii],
                vect_z=pzi - cz[ii],
                strict=False,
                return_x01=True,
            )

            # area
            area[ii] = plg.Polygon(np.array([p0, p1]).T).area()

            # centroid in 3d
            cents0[ii], cents1[ii] = plg.Polygon(
                np.array([x0[ii, :], x1[ii, :]]).T
            ).center()

        centsx, centsy, centsz = coord_x01toxyz(
            x0=cents0,
            x1=cents1,
        )

        plane_nin = coll.dobj[cref][kref]['dgeom']['nin']

    else:
        cents0, cents1 = None, None

    # --------------------
    # reshape if necessary
    # --------------------

    ntot = np.prod(shape0)
    if is2d and harmonize and reshape and x0.shape[0] == ntot:
        shape = tuple(np.r_[shape0, x0.shape[-1]])
        x0 = x0.reshape(shape)
        x1 = x1.reshape(shape)

    # -------------
    # plot & return
    # -------------

    if plot is True:
        out0, out1 = coll.get_optics_outline(key=kref, add_points=False)
        _plot(
            poly_x0=out0,
            poly_x1=out1,
            p0=x0,
            p1=x1,
            cents0=cents0,
            cents1=cents1,
            # options
            tit=f"Reflections on {key}",
            xlab=None,
            ylab=None,
        )

    # timing
    # lt = [f"\t- dt{ii}: {dt[ii]}" for ii in range(len(dt))]
    # print(f"\nTiming for {key}:\n" + "\n".join(lt))

    if return_for_etendue:
        return (
            x0, x1, kref, iok,
            px, py, pz,
            cx, cy, cz,
            centsx, centsy, centsz,
            area, plane_nin, spectro,
        )
    else:
        return x0, x1, kref, iok


# ##############################################################
# ##############################################################
#               check
# ##############################################################


def _check(
    coll=None,
    key=None,
    key_cam=None,
    pixel=None,
    add_points=None,
    convex=None,
    harmonize=None,
    reshape=None,
    verb=None,
    plot=None,
    store=None,
):

    # --------
    # key

    lok = [
        k0 for k0, v0 in coll.dobj.get('diagnostic', {}).items()
        if any([len(v1['optics']) > 0 for v1 in v0['doptics'].values()])
    ]
    key = ds._generic_check._check_var(
        key, 'key',
        types=str,
        allowed=lok,
    )
    spectro = coll.dobj['diagnostic'][key]['spectro']

    # -----------
    # key_cam

    lok =coll.dobj['diagnostic'][key]['camera']
    key_cam = ds._generic_check._check_var(
        key_cam, 'key_cam',
        types=str,
        allowed=lok,
    )

    # --------
    # doptics

    doptics = coll.dobj['diagnostic'][key]['doptics'][key_cam]
    optics = doptics['optics']
    optics_cls = doptics['cls']
    ispectro = doptics.get('ispectro')

    # -------------------------------------------------
    # ldeti: list of individual camera dict (per pixel)

    dgeom = coll.dobj['camera'][key_cam]['dgeom']
    cx, cy, cz = coll.get_camera_cents_xyz(key=key_cam)
    is2d = dgeom['type'] == '2d'
    shape0 = cx.shape

    if is2d:
        cx = cx.ravel()
        cy = cy.ravel()
        cz = cz.ravel()

    # ---------
    # pixel

    if pixel is not None:
        pixel = np.atleast_1d(pixel).astype(int)

        if pixel.ndim == 2 and pixel.shape[1] == 2 and is2d:
            pixel = pixel[:, 0] * shape0[1]  + pixel[:, 1]

        if pixel.ndim != 1:
            msg = "pixel can only have ndim = 2 for 2d cameras!"
            raise Exception(msg)

    # ------------------------------------
    # compute equivalent optics if spectro

    if spectro and len(optics[ispectro[0]+1:]) > 0:

        c0 = (
            len(ispectro) == 1
            or (
                len(ispectro) == 2
                and len(optics[ispectro[1]+1:]) == 0
            )

        )

        # apertures after crystal => reflection
        if c0:
            kref = optics[ispectro[0]]
            cref = optics_cls[ispectro[0]]
            lop_pre = optics[:ispectro[0]]
            lop_pre_cls = optics_cls[:ispectro[0]]
            lop_post = optics[ispectro[0]+1:]
            lop_post_cls = optics_cls[ispectro[0]+1:]

        else:
            raise NotImplementedError()

    else:
        kref = optics[-1]
        cref = optics_cls[-1]
        lop_pre = optics[:-1]
        lop_pre_cls = optics_cls[:-1]
        lop_post = []
        lop_post_cls = []

    # -----------
    # add_points

    add_points = ds._generic_check._check_var(
        add_points, 'add_points',
        types=int,
        default=3,
        sign='>0',
    )

    # -----------
    # convex

    convex = ds._generic_check._check_var(
        convex, 'convex',
        types=bool,
        default=True,
    )

    # -----------
    # harmonnize

    harmonize = ds._generic_check._check_var(
        harmonize, 'harmonize',
        types=bool,
        default=True,
    )

    # -----------
    # reshape

    reshape = ds._generic_check._check_var(
        reshape, 'reshape',
        types=bool,
        default=True,
    )

    # -----------
    # verb

    verb = ds._generic_check._check_var(
        verb, 'verb',
        types=bool,
        default=True,
    )

    # -----------
    # plot

    plot = ds._generic_check._check_var(
        plot, 'plot',
        default=False,
        types=bool,
    )

    # -----------
    # store

    store = ds._generic_check._check_var(
        store, 'store',
        default=False,
        types=bool,
    )

    return (
        key,
        key_cam,
        kref,
        cref,
        spectro,
        ispectro,
        lop_pre,
        lop_pre_cls,
        lop_post,
        lop_post_cls,
        is2d,
        shape0,
        cx,
        cy,
        cz,
        pixel,
        add_points,
        convex,
        harmonize,
        reshape,
        verb,
        plot,
        store,
    )


# ##############################################################
# ##############################################################
#           Equivalent aperture non-spectro
# ##############################################################


def _get_equivalent_aperture(
    p_a=None,
    pt=None,
    nop_pre=None,
    lpoly_pre=None,
    ptsvect=None,
    **kwdargs,
):
    
    # loop on optics
    for jj in range(nop_pre):

        # project on reference frame
        p0, p1 = ptsvect(
            pts_x=pt[0],
            pts_y=pt[1],
            pts_z=pt[2],
            vect_x=lpoly_pre[jj][0] - pt[0],
            vect_y=lpoly_pre[jj][1] - pt[1],
            vect_z=lpoly_pre[jj][2] - pt[2],
            strict=False,
            return_x01=True,
        )[-2:]

        # inside
        if np.all([p_a.isInside(xx, yy) for xx, yy in zip(p0, p1)]):
            p_a = plg.Polygon(np.array([p0, p1]).T)
        else:
            # intersection
            p_a = p_a & plg.Polygon(np.array([p0, p1]).T)
            if p_a.nPoints() < 3:
                return None, None

    # return
    return np.array(p_a.contour(0)).T


# ##############################################################
# ##############################################################
#           Equivalent aperture spectro
# ##############################################################


def _get_equivalent_aperture_spectro(
    p_a=None,
    pt=None,
    nop_pre=None,
    lpoly_pre=None,
    nop_post=None,
    lx0_post=None,
    lx1_post=None,
    # functions
    coord_x01toxyz=None,
    lcoord_x01toxyz_poly=None,
    pts2pt=None,
    ptsvect=None,
    lptsvect_poly=None,
    # options
    add_points=None,
    convex=None,
    # timing
    dt=None,
):

    # loop on optics before crystal
    for jj in range(nop_pre):

        # project on reference frame
        p0, p1 = ptsvect(
            pts_x=pt[0],
            pts_y=pt[1],
            pts_z=pt[2],
            vect_x=lpoly_pre[jj][0] - pt[0],
            vect_y=lpoly_pre[jj][1] - pt[1],
            vect_z=lpoly_pre[jj][2] - pt[2],
            strict=False,
            return_x01=True,
        )[-2:]

        # inside
        if np.all([p_a.isInside(xx, yy) for xx, yy in zip(p0, p1)]):
            p_a = plg.Polygon(np.array([p0, p1]).T)
        else:
            # intersection
            p_a = p_a & plg.Polygon(np.array([p0, p1]).T)
            if p_a.nPoints() < 3:
                return None, None

    # extract p0, p1
    p0, p1 = np.array(p_a.contour(0)).T

    # loop on optics after crystal
    for jj in range(nop_post):
        # print(f'\t {jj} / {nop_post}')      # DB

        # reflection
        p0, p1 = _class5_projections._get_reflection(
            # inital contour
            x0=p0,
            x1=p1,
            # polygon observed
            poly_x0=lx0_post[jj],
            poly_x1=lx1_post[jj],
            # observation point
            pt=pt,
            add_points=add_points,
            # functions
            coord_x01toxyz=coord_x01toxyz,
            coord_x01toxyz_poly=lcoord_x01toxyz_poly[jj],
            pts2pt=pts2pt,
            ptsvect=ptsvect,
            ptsvect_poly=lptsvect_poly[jj],
            # timing
            # dt=dt,
        )

        if p0 is None:
            # print('\t \t None 0')
            return p0, p1

        if np.all([p_a.isInside(xx, yy) for xx, yy in zip(p0, p1)]):
            # print('inside: ', p1)
            # plt.figure()
            # plt.plot(
            #     np.array(p_a.contour(0))[:, 0],
            #     np.array(p_a.contour(0))[:, 1], 
            #     '.-k',
            #     p0, p1, '.-r'
            #     )
            p_a = plg.Polygon(np.array([p0, p1]).T)
        else:
            # convex hull
            if convex:
                p0, p1 = np.array(plgUtils.convexHull(
                    plg.Polygon(np.array([p0, p1]).T)
                ).contour(0)).T

            # plt.figure()
            # plt.plot(
            #     np.array(p_a.contour(0))[:, 0],
            #     np.array(p_a.contour(0))[:, 1], 
            #     '.-k',
            #     p0, p1, '.-r'
            #     )

            # intersection
            p_a = p_a & plg.Polygon(np.array([p0, p1]).T)
            if p_a.nPoints() < 3:
                # print('\t \t None 1')       # DB
                return None, None

            # update
            p0, p1 = np.array(p_a.contour(0)).T
            
            # interpolate
            if jj < nop_post - 1:
                p0, p1 = _compute._interp_poly(
                    lp=[p0, p1],
                    add_points=add_points,
                    mode='min',
                    isclosed=False,
                    closed=False,
                    ravel=True,
                )
                # print(f'\t\t interp => {p0.size} pts')       # DB
            # print('inter: ', p0)

    return p0, p1


# ##############################################################
# ##############################################################
#           Plot
# ##############################################################


def _plot(
    poly_x0=None,
    poly_x1=None,
    p0=None,
    p1=None,
    cents0=None,
    cents1=None,
    # graph options
    fs=None,
    tit=None,
    xlab=None,
    ylab=None,
):

    fig = plt.figure(figsize=fs)
    ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

    ax.set_title(tit)
    ax.set_xlabel(xlab)
    ax.set_ylabel(ylab)

    i0 = np.r_[np.arange(0, poly_x0.size), 0]
    ax.plot(poly_x0[i0], poly_x1[i0], '.-k')
    
    if p0.shape[1] > 0:
        i1 = np.r_[np.arange(0, p0.shape[1]), 0]
        for ii in range(p0.shape[0]):
            ax.plot(p0[ii, i1], p1[ii, i1], '.-', label=f'pix {ii}')

    if cents0 is not None:
        ax.plot(cents0, cents1, 'xr')
        
    ax.legend()
    return



# -*- coding: utf-8 -*-


import warnings
import datetime as dtm


import numpy as np
import matplotlib.pyplot as plt    # DB


import Polygon as plg
import datastock as ds


from . import _class8_compute


# ##############################################################
# ##############################################################
#           Local to global coordinates
# ##############################################################


def _get_reflection(
    # inital contour on crystal
    x0=None,
    x1=None,
    # poly to be observed
    poly_x0=None,
    poly_x1=None,
    # observation point
    pt=None,
    add_points=None,
    # functions
    coord_x01toxyz=None,
    coord_x01toxyz_poly=None,
    pts2pt=None,
    ptsvect=None,
    ptsvect_poly=None,
    # timing
    dt=None,
):

    # -------------------
    #     compute
    # -------------------

    # outline to 3d
    px, py, pz = coord_x01toxyz(x0=x0, x1=x1)

    # Compute poly reflections
    vx, vy, vz, _, iok = ptsvect(
        pts_x=pt[0],
        pts_y=pt[1],
        pts_z=pt[2],
        vect_x=px - pt[0],
        vect_y=py - pt[1],
        vect_z=pz - pt[2],
        strict=False,
        return_x01=False,
    )[3:]

    if not np.any(iok):
        # print('iok')
        return None, None

    # project on target plane
    p0, p1 = ptsvect_poly(
        pts_x=px[iok],
        pts_y=py[iok],
        pts_z=pz[iok],
        vect_x=vx[iok],
        vect_y=vy[iok],
        vect_z=vz[iok],
        strict=False,
        return_x01=True,
    )[-2:]

    pa = plg.Polygon(np.array([poly_x0, poly_x1]).T)

    # isinside
    if np.all([pa.isInside(xx, yy) for xx, yy in zip(p0, p1)]):
        return x0, x1

    # intersection
    p_a = pa & plg.Polygon(np.array([p0, p1]).T)
    if p_a.nPoints() < 3:
        # print('pts < 3')
        return None, None

    # plt.figure()
    # plt.plot(
    #     poly_x0, poly_x1, '.-k',
    #     p0, p1, '.-r',
    #     )

    p0, p1 = np.array(p_a.contour(0)).T

    # interpolate to add points
    if p0.size < 50:
        p0, p1 = _class8_compute._interp_poly(
            lp=[p0, p1],
            add_points=add_points,
            mode='min',
            isclosed=False,
            closed=False,
            ravel=True,
        )

    # back to 3d
    px, py, pz = coord_x01toxyz_poly(x0=p0, x1=p1)

    # back projection on crystal (slowest part)
    p0, p1 = pts2pt(
        pt_x=pt[0],
        pt_y=pt[1],
        pt_z=pt[2],
        # poly
        pts_x=px,
        pts_y=py,
        pts_z=pz,
        # surface
        strict=False,
        return_xyz=False,
        return_x01=True,
        debug=False,
        # timing
        dt=dt,
    )
    
    return p0, p1


# ##############################################################
# ##############################################################
#           Global to local coordinates
# ##############################################################

# DEPRECATED ????
# def _get_project_plane(
#     plane_pt=None,
#     plane_nin=None,
#     plane_e0=None,
#     plane_e1=None,
# ):

#     def _project_poly_on_plane_from_pt(
#         pt_x=None,
#         pt_y=None,
#         pt_z=None,
#         poly_x=None,
#         poly_y=None,
#         poly_z=None,
#         vx=None,
#         vy=None,
#         vz=None,
#         plane_pt=plane_pt,
#         plane_nin=plane_nin,
#         plane_e0=plane_e0,
#         plane_e1=plane_e1,
#     ):

#         sca0 = (
#             (plane_pt[0] - pt_x)*plane_nin[0]
#             + (plane_pt[1] - pt_y)*plane_nin[1]
#             + (plane_pt[2] - pt_z)*plane_nin[2]
#         )

#         if vx is None:
#             vx = poly_x - pt_x
#             vy = poly_y - pt_y
#             vz = poly_z - pt_z

#         sca1 = vx*plane_nin[0] + vy*plane_nin[1] + vz*plane_nin[2]

#         k = sca0 / sca1

#         px = pt_x + k * vx
#         py = pt_y + k * vy
#         pz = pt_z + k * vz

#         p0 = (
#             (px - plane_pt[0])*plane_e0[0]
#             + (py - plane_pt[1])*plane_e0[1]
#             + (pz - plane_pt[2])*plane_e0[2]
#         )
#         p1 = (
#             (px - plane_pt[0])*plane_e1[0]
#             + (py - plane_pt[1])*plane_e1[1]
#             + (pz - plane_pt[2])*plane_e1[2]
#         )

#         return p0, p1

#     def _back_to_3d(
#         x0=None,
#         x1=None,
#         plane_pt=plane_pt,
#         plane_e0=plane_e0,
#         plane_e1=plane_e1,
#     ):

#         return (
#             plane_pt[0] + x0*plane_e0[0] + x1*plane_e1[0],
#             plane_pt[1] + x0*plane_e0[1] + x1*plane_e1[1],
#             plane_pt[2] + x0*plane_e0[2] + x1*plane_e1[2],
#         )

#     return _project_poly_on_plane_from_pt, _back_to_3d



# -*- coding: utf-8 -*-


import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors


import datastock as ds


from ..geom._comp_solidangles import calc_solidangle_apertures


# ##################################################################
# ##################################################################
#             solid angles from any points
# ##################################################################


def compute_solid_angles(
    coll=None,
    key=None,
    key_cam=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # options
    config=None,
    visibility=None,
    # return
    return_vect=None,
    return_alpha=None,
):
    # ---------
    # check

    (
        key, key_cam, spectro,
        ptsx, ptsy, ptsz, shape0_pts,
        return_vect, return_alpha,
    ) = _compute_solid_angles_check(
        coll=coll,
        key=key,
        key_cam=key_cam,
        # pts
        ptsx=ptsx,
        ptsy=ptsy,
        ptsz=ptsz,
        # return
        return_vect=return_vect,
        return_alpha=return_alpha,
    )

    # -----------
    # prepare

    if spectro:
        raise NotImplementedError()

    else:

        dout = _compute_solid_angles_regular(
            coll=coll,
            key=key,
            key_cam=key_cam,
            # pts
            ptsx=ptsx,
            ptsy=ptsy,
            ptsz=ptsz,
            shape0_pts=shape0_pts,
            # options
            config=config,
            visibility=visibility,
            # return
            return_vect=return_vect,
        )

    return dout


def _compute_solid_angles_check(
    coll=None,
    key=None,
    key_cam=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    # options
    config=None,
    visibility=None,
    # return
    return_vect=None,
    return_alpha=None,
):
    # ---------
    # check

    # key_cam
    key, key_cam = coll.get_diagnostic_cam(key=key, key_cam=key_cam)
    spectro = coll.dobj['diagnostic'][key]['spectro']

    # pts
    ptsx = np.atleast_1d(ptsx)
    ptsy = np.atleast_1d(ptsy)
    ptsz = np.atleast_1d(ptsz)

    if not (ptsx.shape == ptsy.shape == ptsz.shape):
        msg = (
            "Args ptsx, ptsy, ptsz must be 3 np.ndarray of the same shape!"
        )
        raise Exception(msg)

    shape0_pts = ptsx.shape
    if ptsx.ndim > 1:
        ptsx = ptsx.ravel()
        ptsy = ptsy.ravel()
        ptsz = ptsz.ravel()

    # return_vect
    return_vect = ds._generic_check._check_var(
        return_vect, 'return_vect',
        types=bool,
        default=False,
    )

    # return_alpha
    return_alpha = ds._generic_check._check_var(
        return_alpha, 'return_alpha',
        types=bool,
        default=False,
    )

    return (
        key, key_cam, spectro,
        ptsx, ptsy, ptsz, shape0_pts,
        return_vect, return_alpha,
    )


def _compute_solid_angles_regular(
    coll=None,
    key=None,
    key_cam=None,
    # pts
    ptsx=None,
    ptsy=None,
    ptsz=None,
    shape0_pts=None,
    # options
    config=None,
    visibility=None,
    # return
    return_vect=None,
):

    doptics = coll.dobj['diagnostic'][key]['doptics']
    dout = {k0: {} for k0 in key_cam}

    for k0 in key_cam:

        # prepare apertures
        dap = {}
        for op, opc in zip(doptics[k0]['optics'], doptics[k0]['cls']):
            dg = coll.dobj[opc][op]['dgeom']
            if dg['type'] == '3d':
                px, py, pz = dg['poly_x'], dg['poly_y'], dg['poly_z']
            else:
                cc = dg['cent']
                out0, out1 = dg['outline']
                out0, out1 = coll.ddata[out0]['data'], coll.ddata[out1]['data']
                px = cc[0] + out0*dg['e0'][0] + out1*dg['e1'][0]
                py = cc[1] + out0*dg['e0'][1] + out1*dg['e1'][1]
                pz = cc[2] + out0*dg['e0'][2] + out1*dg['e1'][2]

            dap[op] = {
                'nin': dg['nin'],
                'poly_x': px,
                'poly_y': py,
                'poly_z': pz,
            }

        # prepare camera
        dg = coll.dobj['camera'][k0]['dgeom']
        ddet = {}

        # cents
        cx, cy, cz = coll.get_camera_cents_xyz(k0)
        npts = cx.size
        sh = cx.shape
        ddet['cents_x'] = cx
        ddet['cents_y'] = cy
        ddet['cents_z'] = cz

        # vectors
        ddet.update(coll.get_camera_unit_vectors(k0))
        for k1 in ['nin', 'e0', 'e1']:
            for ii, ss in enumerate(['x', 'y', 'z']):
                kk = f'{k1}_{ss}'
                if np.isscalar(ddet[kk]):
                    ddet[kk] = np.full(sh, ddet[kk])

        out0, out1 = dg['outline']
        out0, out1 = coll.ddata[out0]['data'], coll.ddata[out1]['data']
        ddet['outline_x0'] = out0
        ddet['outline_x1'] = out1

        # compute
        out = calc_solidangle_apertures(
            # observation points
            pts_x=ptsx,
            pts_y=ptsy,
            pts_z=ptsz,
            # polygons
            apertures=dap,
            detectors=ddet,
            # possible obstacles
            config=config,
            # parameters
            summed=False,
            visibility=visibility,
            return_vector=return_vect,
            return_flat_pts=True,
            return_flat_det=None,
        )

        # store
        if return_vect is True:
            dout[k0]['solid_angle'] = out[0]
            dout[k0]['vectx'] = out[1]
            dout[k0]['vecty'] = out[2]
            dout[k0]['vectz'] = out[3]
        else:
            dout[k0]['solid_angle'] = out

        # reshape
        if shape0_pts != ptsx.shape:
            shape = tuple(np.r_[dout[k0]['solid_angle'].shape, shape0_pts])
            for k1, v1 in dout[k0].items():
                dout[k0][k1] = v1.reshape(shape)

    return dout


# ##################################################################
# ##################################################################
#             interpolated along los
# ##################################################################


def _interpolated_along_los(
    coll=None,
    key=None,
    key_cam=None,
    key_data_x=None,
    key_data_y=None,
    # sampling
    res=None,
    mode=None,
    segment=None,
    radius_max=None,
    # plotting
    vmin=None,
    vmax=None,
    plot=None,
    dcolor=None,
    dax=None,
    ):

    # ------------
    # check inputs

    # key_cam
    key, key_cam = coll.get_diagnostic_cam(key=key, key_cam=key_cam)

    # key_data
    lok_coords = [
        'x', 'y', 'z', 'R', 'phi', 'ang_vs_ephi',
        'k', 'l', 'ltot', 'itot',
    ]
    lok_2d = [
        k0 for k0, v0 in coll.ddata.items()
        if v0.get('bsplines') is not None
    ]

    key_data_x = ds._generic_check._check_var(
        key_data_x, 'key_data_x',
        types=str,
        default='k',
        allowed=lok_coords + lok_2d,
    )

    key_data_y = ds._generic_check._check_var(
        key_data_y, 'key_data_y',
        types=str,
        default='k',
        allowed=lok_coords + lok_2d,
    )

    # segment
    segment = ds._generic_check._check_var(
        segment, 'segment',
        types=int,
        default=-1,
    )

    # plot
    plot = ds._generic_check._check_var(
        plot, 'plot',
        types=bool,
        default=True,
    )

    # dcolor
    if not isinstance(dcolor, dict):
        if dcolor is None:
            lc = ['k', 'r', 'g', 'b', 'm', 'c']
        elif mcolors.is_color_like(dcolor):
            lc = [dcolor]

        dcolor = {
            kk: lc[ii%len(lc)]
            for ii, kk in enumerate(key_cam)
            }

    # --------------
    # prepare output

    ncam = len(key_cam)

    xx = [None for ii in range(ncam)]
    yy = [None for ii in range(ncam)]

    # ---------------
    # loop on cameras

    if key_data_x in lok_coords and key_data_y in lok_coords:

        for ii, kk in enumerate(key_cam):

            klos = coll.dobj['diagnostic'][key]['doptics'][kk]['los']
            if klos is None:
                continue

            xx[ii], yy[ii] = coll.sample_rays(
                key=klos,
                res=res,
                mode=mode,
                segment=segment,
                radius_max=radius_max,
                concatenate=True,
                return_coords=[key_data_x, key_data_y],
                )

            if key_data_x in ['x', 'y', 'z', 'R', 'l', 'ltot']:
                xlab = f"{key_data_x} (m)"
            else:
                xlab = key_data_x

            if key_data_y in ['x', 'y', 'z', 'R', 'l', 'ltot']:
                ylab = f"{key_data_y} (m)"
            else:
                ylab = key_data_y

    elif key_data_x in lok_coords or key_data_y in lok_coords:

        if key_data_x in lok_coords:
            cll = key_data_x
            c2d = key_data_y
            if key_data_x in ['x', 'y', 'z', 'R', 'l', 'ltot']:
                xlab = f"{key_data_x} (m)"
            else:
                xlab = key_data_x
            ylab = f"{key_data_y} ({coll.ddata[key_data_y]['units']})"
        else:
            cll = key_data_y
            c2d = key_data_x
            if key_data_y in ['x', 'y', 'z', 'R', 'l', 'ltot']:
                ylab = f"{key_data_y} (m)"
            else:
                ylab = key_data_y
            xlab = f"{key_data_x} ({coll.ddata[key_data_x]['units']})"

        for ii, kk in enumerate(key_cam):

            klos = coll.dobj['diagnostic'][key]['doptics'][kk]['los']
            if klos is None:
                continue

            pts_x, pts_y, pts_z, pts_ll = coll.sample_rays(
                key=klos,
                res=res,
                mode=mode,
                segment=segment,
                radius_max=radius_max,
                concatenate=True,
                return_coords=['x', 'y', 'z', cll],
                )

            Ri = np.hypot(pts_x, pts_y)

            q2d, _ = coll.interpolate_profile2d(
                key=c2d,
                R=Ri,
                Z=pts_z,
                grid=False,
                crop=True,
                nan0=True,
                val_out=True,
                imshow=False,
                return_params=None,
                store=False,
                inplace=False,
            )

            isok = ~(np.isnan(q2d) & (~np.isnan(Ri)))
            if key_data_x in lok_coords:
                xx[ii] = pts_ll[isok]
                yy[ii] = q2d[isok]
            else:
                xx[ii] = q2d[isok]
                yy[ii] = pts_ll[isok]

    else:
        for ii, kk in enumerate(key_cam):

            klos = coll.dobj['diagnostic'][key]['doptics'][kk]['los']
            if klos is None:
                continue

            pts_x, pts_y, pts_z = coll.sample_rays(
                key=klos,
                res=res,
                mode=mode,
                segment=segment,
                radius_max=radius_max,
                concatenate=True,
                return_coords=['x', 'y', 'z'],
                )

            Ri = np.hypot(pts_x, pts_y)

            q2dx, _ = coll.interpolate_profile2d(
                key=key_data_x,
                R=Ri,
                Z=pts_z,
                grid=False,
                crop=True,
                nan0=True,
                val_out=True,
                imshow=False,
                return_params=None,
                store=False,
                inplace=False,
            )

            q2dy, _ = coll.interpolate_profile2d(
                key=key_data_y,
                R=Ri,
                Z=pts_z,
                grid=False,
                crop=True,
                nan0=True,
                val_out=True,
                imshow=False,
                return_params=None,
                store=False,
                inplace=False,
            )

            isok = ~((np.isnan(q2dx) | np.isnan(q2dy)) & (~np.isnan(Ri)))
            xx[ii] = q2dx[isok]
            yy[ii] = q2dy[isok]

            xlab = f"{key_data_x} ({coll.ddata[key_data_x]['units']})"
            ylab = f"{key_data_y} ({coll.ddata[key_data_y]['units']})"

    # ------------
    # plot

    if plot is True:
        if dax is None:

            fig = plt.figure()

            ax = fig.add_axes([0.1, 0.1, 0.8, 0.8])

            tit = f"{key} LOS"
            ax.set_title(tit, size=12, fontweight='bold')
            ax.set_xlabel(xlab)
            ax.set_ylabel(ylab)

            dax = {'main': ax}

        # main
        kax = 'main'
        if dax.get(kax) is not None:
            ax = dax[kax]

            for ii, kk in enumerate(key_cam):
                ax.plot(
                    xx[ii],
                    yy[ii],
                    c=dcolor[kk],
                    marker='.',
                    ls='-',
                    ms=8,
                    label=kk,
                )

            ax.legend()
            
            if vmin is not None:
                ax.set_ylim(bottom=vmin)
            if vmax is not None:
                ax.set_ylim(top=vmax)

        return xx, yy, dax
    else:
        return xx, yy
